#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import abstractmethod
from collections import UserDict, Counter
from neural_compressor.utils.utility import LazyImport, singleton
from neural_compressor.utils import logger
from neural_compressor.adaptor.pytorch import pytorch_forward_wrapper

import numpy as np

torch = LazyImport('torch')
tf = LazyImport('tensorflow')

@singleton
class TensorflowCriterions(object):
    def __init__(self):
        self.criterions = {}
        self.criterions.update(TENSORFLOW_CRITERIONS)

@singleton
class PyTorchCriterions(object):
    def __init__(self):
        self.criterions = {}
        self.criterions.update(PYTORCH_CRITERIONS)

framework_criterions = {"tensorflow": TensorflowCriterions,
                        "pytorch":    PyTorchCriterions,
                        "pytorch_fx": PyTorchCriterions}

# user/model specific criterions will be registered here
TENSORFLOW_CRITERIONS = {}
PYTORCH_CRITERIONS= {}

registry_criterions = {"tensorflow": TENSORFLOW_CRITERIONS,
                       "pytorch":    PYTORCH_CRITERIONS,
                       "pytorch_fx": PYTORCH_CRITERIONS}

class Criterions(object):
    def __init__(self, framework):
        assert framework in ("tensorflow", "pytorch", "pytorch_fx"), \
                             "framework support tensorflow pytorch"
        self.criterions = framework_criterions[framework]().criterions

    def __getitem__(self, criterion_type):
        assert criterion_type in self.criterions.keys(), "only support criterions in {}".\
            format(self.criterions.keys())

        return self.criterions[criterion_type]

    def register(self, name, criterion_cls):
        assert name not in self.criterions.keys(), 'registered criterion name already exists.'
        self.criterions.update({name: criterion_cls})

def criterion_registry(criterion_type, framework):
    """The class decorator used to register all criterion classes.

    Args:
        criterion_type (str): The string of supported criterion.
        framework (str): The string of supported framework. 

    Returns:
        cls: The class of register.
    """
    def decorator_criterion(cls):
        for fw in [fwk.strip() for fwk in framework.split(',')]:
            assert fw in [
                "tensorflow",
                "pytorch"], "The framework support tensorflow pytorch"

            if criterion_type in registry_criterions[fw].keys():
                raise ValueError('Cannot have two criterions with the same name')
            registry_criterions[fw][criterion_type] = cls
        return cls
    return decorator_criterion

@criterion_registry('CrossEntropyLoss', 'tensorflow')
class TensorFlowCrossEntropyLoss(object):
    """TensorFlow CrossEntropyLoss criterion.

    Args:
        param_dict (dict): The dict of parameters setting by user for CrossEntropyLoss criterion.
    """
    def __init__(self, param_dict):
        assert isinstance(param_dict, dict), 'This criterion constructor parameter must be a dict'
        self._param_dict = param_dict

    def _mapping(self):
        _param_map = {'reduction': 'reduction',
                      'from_logits':'from_logits'}
        _dict = {}
        for key in self._param_dict:
            if key in _param_map:
                if key == 'reduction':
                    assert self._param_dict[key] in ['auto', 'none', 'sum', 'sum_over_batch_size'], \
                        'Supported reduction value for tensorflow is auto, none, sum, sum_over_batch_size'
                _dict.update({_param_map[key]: self._param_dict[key]})
        return _dict

    def __call__(self, **kwargs):
        return tf.keras.losses.CategoricalCrossentropy, self._mapping(**kwargs)

@criterion_registry('SparseCategoricalCrossentropy', 'tensorflow')
class TensorFlowSparseCategoricalCrossentropy(object):
    """TensorFlow SparseCategoricalCrossentropyLoss criterion.

    Args:
        param_dict (dict): The dict of parameters setting by user for SparseCategoricalCrossentropy criterion.
    """
    def __init__(self, param_dict):
        assert isinstance(param_dict, dict), 'This criterion constructor parameter must be a dict'
        self._param_dict = param_dict

    def _mapping(self):
        _param_map = {'reduction': 'reduction',
                      'from_logits':'from_logits'}
        _dict = {}
        for key in self._param_dict:
            if key in _param_map:
                if key == 'reduction':
                    assert self._param_dict[key] in ['auto', 'none', 'sum', 'sum_over_batch_size'], \
                        'Supported reduction value for tensorflow is auto, none, sum, sum_over_batch_size'
                _dict.update({_param_map[key]: self._param_dict[key]})
        return _dict

    def __call__(self, **kwargs):
        return tf.keras.losses.SparseCategoricalCrossentropy, self._mapping(**kwargs)

@criterion_registry('CrossEntropyLoss', 'pytorch')
class PyTorchCrossEntropyLoss(object):
    """PyTorch CrossEntropyLoss criterion.

    Args:
        param_dict (dict): The dict of parameters setting by user for SGD criterion
    """
    def __init__(self, param_dict):
        assert isinstance(param_dict, dict), 'This criterion constructor parameter must be a dict'
        self._param_dict = param_dict

    def _mapping(self):
        _param_map = {'reduction': 'reduction'}
        _dict = {}
        for key in self._param_dict:
            if key in _param_map:
                if key == 'reduction':
                    assert self._param_dict[key] in ['none', 'mean', 'sum'], \
                        'Supported reduction value is none, mean, sum'
                _dict.update({_param_map[key]: self._param_dict[key]})
        return _dict

    def __call__(self, **kwargs):
        return torch.nn.CrossEntropyLoss, self._mapping(**kwargs)

class KnowledgeDistillationFramework(object):
    def __init__(self, student_model=None, teacher_model=None):
        self._student_model = student_model
        self._teacher_model = teacher_model

    @property
    def student_model(self):
        """ Getter of student model """
        return self._student_model

    @student_model.setter
    def student_model(self, model):
        """ Setter of teacher model """
        self._student_model = model

    @property
    def teacher_model(self):
        """ Getter of teacher model """
        return self._teacher_model

    @teacher_model.setter
    def teacher_model(self, model):
        """ Setter of teacher model """
        self._teacher_model = model

class KnowledgeDistillationLoss(KnowledgeDistillationFramework):
    def __init__(self, temperature=1.0, loss_types=['CE', 'CE'], 
                 loss_weights=[0.5, 0.5], student_model=None, teacher_model=None):
        super(KnowledgeDistillationLoss, self).__init__(student_model=student_model,
                                                        teacher_model=teacher_model)
        self.teacher_outputs = None
        self.temperature = temperature
        self.loss_weights = loss_weights
        self.loss_types = loss_types
        self.teacher_student_loss = self.student_targets_loss = None
        assert len(loss_weights) == len(loss_types) == 2, 'Wrong length for ' + \
                                    'loss_weights or loss_types, should be 2.'
        assert sum(loss_weights) == 1.0, 'Sum of loss_weights should be 1.0.'

    def teacher_model_forward(self, input, teacher_model=None):
        raise NotImplementedError('Function teacher_model_forward '
                                  'should be framework related.')

    def teacher_student_loss_cal(self, student_outputs, teacher_outputs):
        raise NotImplementedError('Function teacher_student_loss_cal '
                                  'should be framework related.')

    def student_targets_loss_cal(self, student_outputs, targets):
        raise NotImplementedError('Function student_targets_loss_cal '
                                  'should be framework related.')

    def loss_cal(self, student_outputs, targets):
        return self.student_targets_loss_cal(student_outputs, targets)

    def loss_cal_sloss(self, student_outputs, teacher_outputs, student_loss):
        if self.loss_weights[0] > 0:
            origin_loss = student_loss
        else:
            origin_loss = 0

        if self.loss_weights[1] > 0:
            student_out_ = student_outputs / self.temperature
            teacher_out_ = teacher_outputs / self.temperature
            distillation_loss = self.teacher_student_loss_cal(student_out_, teacher_out_)
            distillation_loss *= self.temperature ** 2
        else:
            distillation_loss = 0

        self.loss = origin_loss * self.loss_weights[0] + \
                    distillation_loss * self.loss_weights[1]
        return self.loss

    def __call__(self, student_outputs, targets):
        return self.loss_cal(student_outputs, targets)

class PyTorchKnowledgeDistillationLoss(KnowledgeDistillationLoss):
    def __init__(self, temperature=1.0, loss_types=['CE', 'CE'], 
                 loss_weights=[0.5, 0.5], student_model=None, teacher_model=None):
        super(PyTorchKnowledgeDistillationLoss, self).__init__(temperature=temperature, 
                                                               loss_types=loss_types,
                                                               loss_weights=loss_weights,
                                                               student_model=student_model,
                                                               teacher_model=teacher_model)
        if self.student_targets_loss is None:
            if self.loss_types[0] == 'CE':
                self.student_targets_loss = torch.nn.CrossEntropyLoss()
            elif self.loss_types[0] == 'MSE':
                self.student_targets_loss = torch.nn.MSELoss()
            else:
                raise NotImplementedError('Now we only support CrossEntropyLoss and MSELoss '
                 'for loss of student model output with respect to targets.')
            logger.info('student_targets_loss: {}, {}'.format(self.loss_types[0], \
                                                        self.loss_weights[0]))
        if self.teacher_student_loss is None:
            if self.loss_types[1] == 'CE':
                self.teacher_student_loss = self.SoftCrossEntropy
            elif self.loss_types[1] == 'KL':
                self.teacher_student_loss = self.KullbackLeiblerDivergence
            elif self.loss_types[1] == 'MSE':
                self.teacher_student_loss = torch.nn.MSELoss()
            else:
                raise NotImplementedError('Now we only support CrossEntropyLoss KL Divergence'
                ' and MSELoss for loss of student model output with respect to teacher model ouput.')
            logger.info('teacher_student_loss: {}, {}'.format(self.loss_types[1], \
                                                        self.loss_weights[1]))

    def SoftCrossEntropy(self, logits, targets):
        log_prob = torch.nn.functional.log_softmax(logits, dim=-1)
        targets_prob = torch.nn.functional.softmax(targets, dim=-1)
        return (- targets_prob * log_prob).sum(dim=-1).mean()

    def KullbackLeiblerDivergence(self, logits, targets):
        log_prob = torch.nn.functional.log_softmax(logits, dim=-1)
        targets_prob = torch.nn.functional.softmax(targets, dim=-1)
        return torch.nn.functional.kl_div(log_prob, targets_prob)

    def teacher_model_forward(self, input, teacher_model=None, device=None):
        outputs = None
        if self.loss_weights[1] > 0:
            model = self.teacher_model if teacher_model is None else teacher_model
            assert isinstance(model, torch.nn.Module), \
                'Teacher model should be a torch Module instead of {}'.format(type(model))
            model.eval()
            try:
                model_device = next(model.parameters()).device
            except:
                logger.warning("Cannot get model device, assuming it's in CPU.")
                model_device = "cpu"
            device = model_device if device is None else device
            if device != model_device:
                model.to(device)
            with torch.no_grad():
                outputs = pytorch_forward_wrapper(model, input, device=device)
            self.teacher_outputs = outputs
        return outputs

    def teacher_student_loss_cal(self, student_outputs, teacher_outputs):
        assert self.teacher_student_loss, 'teacher_student_loss not specified.'
        return self.teacher_student_loss(student_outputs, teacher_outputs)

    def student_targets_loss_cal(self, student_outputs, targets):
        assert self.student_targets_loss, 'student_targets_loss not specified.'
        return self.student_targets_loss(student_outputs, targets)

@criterion_registry('KnowledgeDistillationLoss', 'pytorch')
class PyTorchKnowledgeDistillationLossWrapper(object):
    def __init__(self, param_dict):
        self.param_dict = param_dict

    def _param_check(self):
        param_dict = self.param_dict
        _params = ['temperature', 'loss_types', 'loss_weights']
        assert all(key in param_dict for key in _params),\
            'Keys {} must be in input parameters.'.format(_params)
        assert param_dict['temperature'] > 0.0,\
            'Value of temperature must be positive.'
        assert len(param_dict['loss_types']) == len(param_dict['loss_weights']),\
            'Length of loss_types and loss_weights must be the same.'
        assert all(type(param_dict[k]) in [list, tuple] \
            for k in ['loss_types', 'loss_weights']),\
            'Type of loss_types and loss_weights must be list or tuple.'
        assert all(any(isinstance(e, t) for t in [str, torch.nn.Module]) \
            for e in param_dict['loss_types']), \
            'Type of loss_types element must be str or torch Module.'
        assert all(0. <= e <= 1. for e in param_dict['loss_weights']) and \
            abs(sum(param_dict['loss_weights']) - 1.0) < 1e-9, \
            'Element of loss_weights must be in interval [0, 1] and summed to 1.0.'
        new_dict = {}
        for k in _params:
            new_dict[k] = param_dict[k]
        return new_dict

    def __call__(self, **kwargs):
        return PyTorchKnowledgeDistillationLoss, self._param_check()

class TensorflowKnowledgeDistillationLoss(KnowledgeDistillationLoss):
    def __init__(self, temperature=1.0, loss_types=['CE', 'CE'], 
                 loss_weights=[0.5, 0.5], student_model=None, teacher_model=None):
        super(TensorflowKnowledgeDistillationLoss, self).__init__(temperature=temperature, 
                                                                  loss_types=loss_types,
                                                                  loss_weights=loss_weights,
                                                                  student_model=student_model,
                                                                  teacher_model=teacher_model)
        if self.student_targets_loss is None:
            if self.loss_types[0] == 'CE':
                self.student_targets_loss = tf.keras.losses.SparseCategoricalCrossentropy()
            else:
                raise NotImplementedError('Now we only support CrossEntropyLoss '
                 'for loss of student model output with respect to targets.')
            logger.info('student_targets_loss: {}, {}'.format(self.loss_types[0], \
                                                        self.loss_weights[0]))            
        if self.teacher_student_loss is None:
            if self.loss_types[1] == 'CE':
                self.teacher_student_loss = self.SoftCrossEntropy
            elif self.loss_types[1] == 'KL':
                self.teacher_student_loss = tf.keras.losses.KLDivergence()
            else:
                raise NotImplementedError('Now we only support CrossEntropyLoss'
                ' for loss of student model output with respect to teacher model ouput.')
            logger.info('teacher_student_loss: {}, {}'.format(self.loss_types[1], \
                                                        self.loss_weights[1]))
    def SoftCrossEntropy(self, targets, logits):
        log_prob = tf.math.log(logits)
        targets_prob = targets
        return tf.math.reduce_mean(tf.math.reduce_sum(- targets_prob * log_prob, axis=-1), axis=-1)

    def teacher_model_forward(self, input, teacher_model=None):
        outputs = None
        if self.loss_weights[1] > 0 and input is not None:
            model = self.teacher_model if teacher_model is None else teacher_model
            if isinstance(input, list) or isinstance(input, tuple):  # pragma: no cover
                outputs = model(*input, training=True)
            elif isinstance(input, dict):  # pragma: no cover
                outputs = model(**input, training=True)
            else:
                outputs = model(input, training=True)
            self.teacher_outputs = outputs
        return outputs

    def teacher_student_loss_cal(self, student_outputs, teacher_outputs):
        assert self.teacher_student_loss, 'teacher_student_loss not specified.'
        return self.teacher_student_loss(teacher_outputs, student_outputs)

    def student_targets_loss_cal(self, student_outputs, targets):
        assert self.student_targets_loss, 'student_targets_loss not specified.'
        return self.student_targets_loss(targets, student_outputs)

    def __call__(self, student_outputs, targets):
        tmp = student_outputs
        student_outputs = targets
        targets = tmp
        return self.loss_cal(student_outputs, targets)

@criterion_registry('KnowledgeDistillationLoss', 'tensorflow')
class TensorflowKnowledgeDistillationLossWrapper(object):
    def __init__(self, param_dict):
        self.param_dict = param_dict

    def _param_check(self):
        param_dict = self.param_dict
        _params = ['temperature', 'loss_types', 'loss_weights']
        assert all(key in param_dict for key in _params),\
            'Keys {} must be in input parameters.'.format(_params)
        assert param_dict['temperature'] > 0.0,\
            'Value of temperature must be positive.'
        assert len(param_dict['loss_types']) == len(param_dict['loss_weights']),\
            'Length of loss_types and loss_weights must be the same.'
        assert all(type(param_dict[k]) in [list, tuple] \
            for k in ['loss_types', 'loss_weights']),\
            'Type of loss_types and loss_weights must be list or tuple.'
        assert all(any(isinstance(e, t) for t in [str, tf.keras]) \
            for e in param_dict['loss_types']), \
            'Type of loss_types element must be str or torch Module.'
        assert all(0. <= e <= 1. for e in param_dict['loss_weights']) and \
            abs(sum(param_dict['loss_weights']) - 1.0) < 1e-9, \
            'Element of loss_weights must be in interval [0, 1] and summed to 1.0.'
        new_dict = {}
        for k in _params:
            new_dict[k] = param_dict[k]
        return new_dict

    def __call__(self, **kwargs):
        return TensorflowKnowledgeDistillationLoss, self._param_check()

class TensorflowKnowledgeDistillationLossExternal(KnowledgeDistillationLoss):
    def __init__(self, temperature=1.0, loss_types=['CE', 'CE'], 
                 loss_weights=[0.5, 0.5], student_model=None, teacher_model=None):
        super(TensorflowKnowledgeDistillationLossExternal, self).__init__(
                                                                temperature=temperature, 
                                                                loss_types=loss_types,
                                                                loss_weights=loss_weights,
                                                                student_model=student_model,
                                                                teacher_model=teacher_model)
        if self.student_targets_loss is None:
            if self.loss_types[0] == 'CE':
                self.student_targets_loss = tf.keras.losses.CategoricalCrossentropy()
            else:
                raise NotImplementedError('Now we only support CrossEntropyLoss '
                 'for loss of student model output with respect to targets.')
            logger.info('student_targets_loss: {}, {}'.format(self.loss_types[0], \
                                                        self.loss_weights[0]))            
        if self.teacher_student_loss is None:
            if self.loss_types[1] == 'CE':
                self.teacher_student_loss = tf.keras.losses.CategoricalCrossentropy()
            elif self.loss_types[1] == 'KL':
                self.teacher_student_loss = tf.keras.losses.KLDivergence()
            else:
                raise NotImplementedError('Now we only support CrossEntropyLoss'
                ' for loss of student model output with respect to teacher model ouput.')
            logger.info('teacher_student_loss: {}, {}'.format(self.loss_types[1], \
                                                        self.loss_weights[1]))

    def teacher_model_forward(self, input, teacher_model=None):
        outputs = None
        if self.loss_weights[1] > 0 and input is not None:
            model = self.teacher_model if teacher_model is None else teacher_model
            if isinstance(input, list) or isinstance(input, tuple):  # pragma: no cover
                outputs = model(*input, training=True)
            elif isinstance(input, dict):  # pragma: no cover
                outputs = model(**input, training=True)
            else:
                outputs = model(input, training=True)
            self.teacher_outputs = outputs
        return outputs

    def teacher_student_loss_cal(self, student_outputs, teacher_outputs):
        assert self.teacher_student_loss, 'teacher_student_loss not specified.'
        return self.teacher_student_loss(teacher_outputs, student_outputs)

    def student_targets_loss_cal(self, student_outputs, targets):
        assert self.student_targets_loss, 'student_targets_loss not specified.'
        return self.student_targets_loss(targets, student_outputs)


class IntermediateLayersKnowledgeDistillationLoss(KnowledgeDistillationFramework):
    def __init__(self, layer_mappings=[], loss_types=None, loss_weights=None, 
                 add_origin_loss=False, student_model=None, teacher_model=None):
        super(IntermediateLayersKnowledgeDistillationLoss, self).__init__(
            student_model=student_model,
            teacher_model=teacher_model
            )
        self.student_features = {}
        self.teacher_features = {}

        self.layer_mappings = []
        self.layer_output_process = []
        for item in layer_mappings:
            assert len(item) == 4 or len(item) == 2, 'Each item in layer_mappings ' + \
                'should be a list or tuple of length 2 or 4, with format ' + \
                '[student_layer_name, teacher_layer_name] or ' + \
                '[student_layer_name, student_layer_output_process, ' + \
                'teacher_layer_name, teacher_layer_output_process].' + \
                'For example, with length of 4, element looks like [\'student_model.layer1' + \
                '.attention\', \'1\', \'teacher_model.layer1.attention\', \'1\'], where ' + \
                '\'student_model.layer1.attention\' and \'teacher_model.layer1.attention\' ' + \
                'represent attention module on layer 1 of the student model and the ' + \
                'teacher model respectively, two \'1\' represent the index to retrieve the ' + \
                'desired output from the defined module\'s outputs, in this case, the above ' + \
                'two module\'s outputs are lists, with desired output in index 1 of these ' + \
                'lists, in cases of dict output, retrieving can be done by define the ' + \
                'corresponding key, in cases of module\'s output is the desired output, ' + \
                'just adopt an element of length 2 such as [\'student_model.layer1.output' + \
                '.output\', \'teacher_model.layer1.output\'].'
            self.layer_mappings.append((item[0], item[2]) if len(item)==4 else item)
            self.layer_output_process.append((item[1], item[3]) if len(item)==4 else ('', ''))
        for student_layer, teacher_layer in self.layer_mappings:
            self.student_features[student_layer] = []
            self.teacher_features[teacher_layer] = []

        self.loss_weights = [1.0 / len(layer_mappings)] * len(layer_mappings) \
                            if loss_weights is None else loss_weights
        self.loss_types = ['MSE'] * len(layer_mappings) \
                          if loss_types is None else loss_types
        self.add_origin_loss = add_origin_loss
        self.loss_funcs = []
        self.feature_matchers = None
        self.init_loss_funcs()
        assert len(self.layer_mappings) == len(self.loss_weights) == len(self.loss_types), \
            'Wrong length for layer_mappings:{}, loss_weights:{} or loss_types:{}, ' + \
            'all should be the same.'.format(
                len(self.layer_mappings), len(self.loss_weights), len(self.loss_types)
                )

    def init_loss_funcs(self):
        raise NotImplementedError('Function init_loss_funcs '
                                  'should be framework related.')

    def init_feature_matcher(self, student_feature, teacher_feature):
        raise NotImplementedError('Function init_feature_matcher '
                                  'should be framework related.')

    def teacher_model_forward(self, input, teacher_model=None):
        raise NotImplementedError('Function teacher_model_forward '
                                  'should be framework related.')

    def loss_cal(self):
        raise NotImplementedError('Function loss_cal should be framework related.')

    def loss_cal_sloss(self, student_outputs, teacher_outputs, student_loss):
        return self.loss_cal()

    def clear_features(self):
        for student_layer in self.student_features:
            self.student_features[student_layer] = []
        for teacher_layer in self.teacher_features:
            self.teacher_features[teacher_layer] = []

    def __call__(self, student_outputs, targets):
        return 0


class PyTorchIntermediateLayersKnowledgeDistillationLoss(
                IntermediateLayersKnowledgeDistillationLoss
                ):
    def __init__(self, layer_mappings=[], loss_types=None, loss_weights=None, 
                 add_origin_loss=False, student_model=None, teacher_model=None):
        super(PyTorchIntermediateLayersKnowledgeDistillationLoss, self).__init__(
                                                layer_mappings=layer_mappings, 
                                                loss_types=loss_types,
                                                loss_weights=loss_weights,
                                                add_origin_loss=add_origin_loss, 
                                                student_model=student_model,
                                                teacher_model=teacher_model)
        self.register_hooks_for_models()

    def register_hooks_for_models(self):
        from neural_compressor.experimental.common import torch_utils
        def register_model_forward_hook(model, path, output_process='', student=False):
            nodes = path.split('.')
            module = model
            for node in nodes:
                try:
                    module = module.__getattr__(node)
                except:
                    raise AttributeError('There is no path {} in the model.'.format(path))
            return module.register_forward_hook(
                torch_utils.get_activation(path, output_process, student)
            )

        assert isinstance(self.student_model, torch.nn.Module) and \
               isinstance(self.teacher_model, torch.nn.Module), \
               'Expect student_model and teacher_model to be an torch.nn.Module object, ' + \
               'got student_model:{} and teacher_model:{}'.format(
                   type(self.student_model), type(self.teacher_model)
                   )
        self.hook_handles = []
        for idx in range(len(self.layer_mappings)):
            student_layer, teacher_layer = self.layer_mappings[idx]
            student_output_process, teacher_output_process = self.layer_output_process[idx]
            st_handle = register_model_forward_hook(self.student_model, student_layer,
                                                    student_output_process, True)
            te_handle = register_model_forward_hook(self.teacher_model, teacher_layer,
                                                    teacher_output_process)
            torch_utils.STUDENT_FEATURES = self.student_features
            torch_utils.TEACHER_FEATURES = self.teacher_features
            self.hook_handles.extend([st_handle, te_handle])

    def remove_all_hooks(self):
        for hook in self.hook_handles:
            hook.remove()

    def init_loss_funcs(self):
        for loss_type in self.loss_types:
            if loss_type == 'MSE':
                loss_func = torch.nn.MSELoss()
            elif loss_type == 'KL':
                loss_func = torch.nn.KLDivLoss()
            elif loss_type == 'L1':
                loss_func = torch.nn.L1Loss()
            else:
                raise NotImplementedError('Unsupported loss type {}, supported loss is ' + \
                    'MSE for mean squared error, KL for Kullback-Leibler divergence and ' + \
                    'L1 for L1 loss.'.format(loss_type))
            self.loss_funcs.append(loss_func)

    def init_feature_matcher(self, student_feature, teacher_feature):
        class pytorch_linear_feature_matcher(torch.nn.Module):
            def __init__(self, src_shape, dst_shape):
                super().__init__()
                shape_diff = [abs(i - j) for i, j in zip(dst_shape, src_shape)]
                assert shape_diff.count(0) == len(shape_diff) - 1, 'Expect only one ' + \
                        'different dimension between student_feature and teacher_feature.'
                self.dim_idx = np.argmax(shape_diff)
                self.dense = torch.nn.Linear(src_shape[self.dim_idx], dst_shape[self.dim_idx])

            def forward(self, input):
                output = torch.transpose(input, self.dim_idx, -1)
                if input.device != next(self.parameters()).device:
                    self.to(input.device)
                output = self.dense(output)
                output = torch.transpose(output, self.dim_idx, -1)
                return output

        assert isinstance(student_feature, (torch.Tensor, np.ndarray)) and \
            isinstance(teacher_feature, (torch.Tensor, np.ndarray)), \
            'Expect student_feature and teacher_feature to be torch.Tensor or np.ndarray ' + \
            'objects, got student_feature a {st} object, teacher_feature a {tt} object.'.format(
                st=type(student_feature), tt=type(teacher_feature)
            )
        assert len(student_feature.shape) == len(teacher_feature.shape), \
            'Expect student_feature and teacher_feature to have the same length of shape, ' + \
            'got student_feature of {}, teacher_feature of {}.'.format(
                student_feature.shape, teacher_feature.shape
                )
        if sum([abs(i - j) for i, j in zip(student_feature.shape, teacher_feature.shape)]) == 0:
            return lambda x:x
        return pytorch_linear_feature_matcher(student_feature.shape, teacher_feature.shape)

    def teacher_model_forward(self, input, teacher_model=None, device=None):
        model = self.teacher_model if teacher_model is None else teacher_model
        assert isinstance(model, torch.nn.Module), \
            'Teacher model should be a torch Module instead of {}'.format(type(model))
        model.eval()
        try:
            model_device = next(model.parameters()).device
        except:
            logger.warning("Cannot get model device, assuming it's in CPU.")
            model_device = "cpu"
        device = model_device if device is None else device
        if device != model_device:
            model.to(device)
        with torch.no_grad():
            return pytorch_forward_wrapper(model, input, device=device)

    def loss_cal_sloss(self, student_outputs, teacher_outputs, student_loss):
        loss = self.loss_cal()
        if self.add_origin_loss:
            loss += student_loss
        return loss

    def loss_cal(self):
        self.loss = 0
        init_feature_matchers = False
        if self.feature_matchers is None:
            init_feature_matchers = True
            self.feature_matchers = {}
        for idx in range(len(self.layer_mappings)):
            student_layer, teacher_layer = self.layer_mappings[idx]
            student_feature = self.student_features[student_layer]
            teacher_feature = self.teacher_features[teacher_layer]
            assert len(student_feature) == len(teacher_feature) and len(student_feature) > 0, \
                'Lengths of student_feature and teacher_feature should be the same and larger than 0, ' + \
                'instead of {} and {}, '.format(len(student_feature), len(teacher_feature)) + \
                'please run student and teacher model forward properly before calculating the loss.'
            def device2feature_gen(features):
                devices_count = Counter([f.device for f in features])
                assert [1] * len(devices_count) == [_ for _ in devices_count.values()], \
                    'Currently only support 1 feature tensor per device, ' + \
                    'got {}.'.format(devices_count)
                return {feat.device:feat for feat in features}

            student_feature = device2feature_gen(student_feature)
            teacher_feature = device2feature_gen(teacher_feature)
            assert student_feature.keys() == teacher_feature.keys(), \
                'Features from student model have different devices with that of ' + \
                'teacher model, got student: {}, teacher: {}.'.format(student_feature.keys(),
                                                                    teacher_feature.keys())
            output_device = torch.device('cuda:0') \
                if torch.device('cuda:0') in student_feature.keys() else torch.device('cpu')
            if init_feature_matchers:
                feature_matcher = self.init_feature_matcher(student_feature[output_device], 
                                                            teacher_feature[output_device])
                self.feature_matchers[student_layer] = feature_matcher

            tmp_loss = 0
            for device in student_feature.keys():
                student_feature[device] = student_feature[device].to(output_device)
                teacher_feature[device] = teacher_feature[device].to(output_device)
                stfeat, tefeat = student_feature[device], teacher_feature[device]
                stfeat = self.feature_matchers[student_layer](stfeat)
                if self.loss_types[idx] == 'KL':
                    check_is_not_prob = \
                        lambda x:(torch.abs(x.sum(dim=-1) - 1.0) > 0.2).any().item()
                    if isinstance(self.feature_matchers[student_layer], torch.nn.Module):
                        stfeat = torch.nn.LogSoftmax(dim=-1)(stfeat)
                    else:
                        if check_is_not_prob(stfeat):
                            stfeat = torch.softmax(stfeat, dim=-1)
                        stfeat = torch.log(stfeat+1e-9)
                    if check_is_not_prob(tefeat):
                        tefeat = torch.softmax(tefeat, dim=-1)
                tmp_loss += self.loss_funcs[idx](stfeat, tefeat) * self.loss_weights[idx]
            self.loss += tmp_loss
        self.clear_features()
        return self.loss

@criterion_registry('IntermediateLayersKnowledgeDistillationLoss', 'pytorch')
class PyTorchIntermediateLayersKnowledgeDistillationLossWrapper(object):
    def __init__(self, param_dict):
        self.param_dict = param_dict

    def _param_check(self):
        param_dict = self.param_dict
        _params = ['layer_mappings', 'loss_types', 'loss_weights', 'add_origin_loss']
        layer_mappings = param_dict['layer_mappings']
        if 'loss_types' not in param_dict:
            param_dict['loss_types'] = ['MSE'] * len(layer_mappings)
        if 'loss_weights' not in param_dict:
            param_dict['loss_weights'] = [1.0 / len(layer_mappings)] * len(layer_mappings)
        if 'add_origin_loss' not in param_dict:
            param_dict['add_origin_loss'] = False
        assert 'layer_mappings' in param_dict, \
            'Key layer_mappings must be in input parameters.'
        assert all(type(param_dict[k]) in [list, tuple] \
            for k in ['layer_mappings', 'loss_types', 'loss_weights']), \
            'Type of loss_types and loss_weights must be list or tuple.'
        assert isinstance(param_dict['add_origin_loss'], bool), \
            'Type of add_origin_loss should be bool.'
        assert len(param_dict['layer_mappings']) == \
            len(param_dict['loss_types']) == len(param_dict['loss_weights']),\
            'Length of layer_mappings, loss_types and loss_weights must be the same.'
        assert all(type(it) in [list, tuple] and (len(it) == 2 or len(it) == 4) \
            for it in param_dict['layer_mappings']), \
            'Elements of layer_mappings must be list or tuple and with length of 2 or 4.' + \
            'For example, with length of 4, element looks like [\'student_model.layer1' + \
            '.attention\', \'1\', \'teacher_model.layer1.attention\', \'1\'], where ' + \
            '\'student_model.layer1.attention\' and \'teacher_model.layer1.attention\' ' + \
            'represent attention module on layer 1 of the student model and the ' + \
            'teacher model respectively, two \'1\' represent the index to retrieve the ' + \
            'desired output from the defined module\'s outputs, in this case, the above ' + \
            'two module\'s outputs are lists, with desired output in index 1 of these ' + \
            'lists, in cases of dict output, retrieving can be done by define the ' + \
            'corresponding key, in cases of module\'s output is the desired output, ' + \
            'just adopt an element of length 2 such as [\'student_model.layer1.output' + \
            '.output\', \'teacher_model.layer1.output\'].'
        assert all(any(isinstance(e, t) for t in [str, torch.nn.Module]) \
            for e in param_dict['loss_types']), \
            'Type of loss_types element must be str or torch Module.'
        assert all(0. <= e <= 1. for e in param_dict['loss_weights']), \
            'Element of loss_weights must be in interval [0, 1].'
        new_dict = {}
        for k in _params:
            new_dict[k] = param_dict[k]
        return new_dict

    def __call__(self, **kwargs):
        return PyTorchIntermediateLayersKnowledgeDistillationLoss, self._param_check()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import abstractmethod
from neural_compressor.utils.create_obj_from_config import get_algorithm

registry_algorithms = {}

def algorithm_registry(algorithm_type):
    """The class decorator used to register all Algorithm subclasses.

    Args:
        cls (class): The class of register.
        algorithm_type (str): The algorithm registration name

    Returns:
        cls: The class of register.
    """
    def decorator_algorithm(cls):
        if algorithm_type in registry_algorithms:
            raise ValueError('Cannot have two algorithms with the same name')
        registry_algorithms[algorithm_type] = cls
        return cls
    return decorator_algorithm

class ALGORITHMS(object):
    algorithms = registry_algorithms

    def __getitem__(self, algorithm_type):
        assert algorithm_type in self.algorithms, "algorithm type only support {}".\
            format(self.algorithms.keys())
        return self.algorithms[algorithm_type]

    @classmethod
    def support_algorithms(self):
        return set(self.algorithms.keys())

class AlgorithmScheduler(object):
    """control the Algorithm in different phase

    """
    def __init__(self, conf):
        self.algorithms = get_algorithm(ALGORITHMS, conf)
        self._origin_model = None
        self._q_model = None
        self._dataloader = None
        self._adaptor = None
        self._calib_iter = None

    def __call__(self):
        assert self._q_model, 'set q_model for algorithm'
        if len(self.algorithms) == 0:
            return self._q_model
        assert self._origin_model, 'set origin model for algorithm'
        assert self._dataloader, 'set dataloader for algorithm'
        assert self._adaptor, 'set adaptor for algorithm'
        assert self._calib_iter, 'set calibration iteration for algorithm'
        for algo in self.algorithms:
            self._q_model = algo(self._origin_model, 
                                 self._q_model, \
                                 self._adaptor, \
                                 self._dataloader, \
                                 self._calib_iter)
        return self._q_model

    @property
    def origin_model(self):
        return self._origin_model

    @origin_model.setter
    def origin_model(self, model):
        self._origin_model = model

    @property
    def q_model(self):
        return self._q_model

    @q_model.setter
    def q_model(self, model):
        self._q_model = model

    @property
    def dataloader(self):
        return self._dataloader

    @dataloader.setter
    def dataloader(self, dataloader):
        self._dataloader = dataloader

    @property
    def adaptor(self):
        return self._adaptor

    @adaptor.setter
    def adaptor(self, adaptor):
        self._adaptor = adaptor

    @property
    def calib_iter(self):
        return self._calib_iter

    @calib_iter.setter
    def calib_iter(self, calib_iter):
        self._calib_iter = calib_iter

class Algorithm(object):
    """ The base class of algorithm. 

    """
    @abstractmethod
    def __call__(self, *args, **kwargs):
        raise NotImplementedError




# Copyright (c) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# from reportlab.pdfgen import canvas
# from reportlab.platypus import (
#     SimpleDocTemplate, Paragraph, PageBreak, Image, Spacer, Table, TableStyle)
# from reportlab.lib.enums import TA_LEFT, TA_RIGHT, TA_CENTER, TA_JUSTIFY
# from reportlab.lib.styles import ParagraphStyle, getSampleStyleSheet
# from reportlab.lib.pagesizes import LETTER, inch
# from reportlab.graphics.shapes import Line, LineShape, Drawing, String
# from reportlab.lib.colors import Color


# class FooterCanvas(canvas.Canvas):

#     def __init__(self, *args, **kwargs):
#         canvas.Canvas.__init__(self, *args, **kwargs)
#         self.pages = []
#         self.width, self.height = LETTER

#     def showPage(self):
#         self.pages.append(dict(self.__dict__))
#         self._startPage()

#     def save(self):
#         page_count = len(self.pages)
#         for page in self.pages:
#             self.__dict__.update(page)
#             if (self._pageNumber > 1):
#                 self.draw_canvas(page_count)
#             canvas.Canvas.showPage(self)
#         canvas.Canvas.save(self)

#     def draw_canvas(self, page_count):
#         page = "Page %s of %s" % (self._pageNumber, page_count)
#         x = 128
#         self.saveState()
#         self.setStrokeColorRGB(0, 0, 0)
#         self.setLineWidth(0.5)
#         self.line(20, 740, LETTER[0] - 50, 740)
#         self.line(66, 78, LETTER[0] - 66, 78)
#         self.setFont('Times-Roman', 10)
#         self.drawString(LETTER[0] - x, 65, page)
#         self.restoreState()


# class PDFReport:

#     def __init__(self,
#                  path,
#                  list_optimization_set_top3,
#                  list_performance_top3,
#                  original_model_ranking,
#                  original_model_performance,
#                  list_config_best_ncpi,
#                  list_config_best_nins,
#                  list_config_best_bs,
#                  list_config_best_performance,
#                  TCO_unit_pricing,  # 2.448
#                  cloud_vendor,  # "AWS"
#                  cloud_instance_type,  # "c6i"
#                  ):
#         self.path = path
#         self.styleSheet = getSampleStyleSheet()
#         self.elements = []

#         # from superbench results
#         self.list_optimization_set_top3 = list_optimization_set_top3
#         self.list_performance_top3 = list_performance_top3
#         self.original_model_ranking = original_model_ranking
#         self.original_model_performance = original_model_performance

#         self.list_config_best_ncpi = list_config_best_ncpi
#         self.list_config_best_nins = list_config_best_nins
#         self.list_config_best_bs = list_config_best_bs
#         self.list_config_best_performance = list_config_best_performance

#         # from Cloud
#         self.TCO_unit_pricing = TCO_unit_pricing
#         self.cloud_vendor = cloud_vendor
#         self.cloud_instance_type = cloud_instance_type

#         # colors
#         self.IntelBlueDark = Color((0 / 255), (74 / 255), (134 / 255), 1)
#         self.IntelBlueLight = Color((0 / 255), (178 / 255), (227 / 255), 1)
#         self.IntelBlueReg = Color((0 / 255), (104 / 255), (181 / 255), 1)

#         self.nextPagesHeader(True)
#         self.MainpageMaker()

#         self.doc = SimpleDocTemplate(path, pagesize=LETTER)
#         self.doc.multiBuild(self.elements, canvasmaker=FooterCanvas)

#     def nextPagesHeader(self, isSecondPage):
#         if isSecondPage:
#             psHeaderText = ParagraphStyle(
#                 'Hed0', fontSize=20, alignment=TA_LEFT, borderWidth=3, textColor=self.IntelBlueDark)
#             text = '<b>AI Deployment Report</b>'
#             paragraphReportHeader = Paragraph(text, psHeaderText)
#             self.elements.append(paragraphReportHeader)

#             spacer = Spacer(10, 15)
#             self.elements.append(spacer)

#             d = Drawing(500, 1)
#             line = Line(0, 0, 460, 0)
#             line.strokeColor = self.IntelBlueReg
#             line.strokeWidth = 2.5
#             d.add(line)
#             self.elements.append(d)

#             spacer = Spacer(10, 2)
#             self.elements.append(spacer)

#             d = Drawing(500, 1)
#             line = Line(0, 0, 460, 0)
#             line.strokeColor = self.IntelBlueReg
#             line.strokeWidth = 1
#             d.add(line)
#             self.elements.append(d)

#             spacer = Spacer(10, 15)
#             self.elements.append(spacer)

#     def MainpageMaker(self):
#         # Optimization Table

#         psHeaderText = ParagraphStyle(
#             'Hed0', fontSize=16, alignment=TA_LEFT, borderWidth=3, textColor=self.IntelBlueDark)
#         text = '<b>Best Optimization (Default Config)</b>'
#         paragraphReportHeader = Paragraph(text, psHeaderText)
#         self.elements.append(paragraphReportHeader)

#         spacer = Spacer(5, 11)
#         self.elements.append(spacer)

#         psHeaderText = ParagraphStyle(
#             'Hed0', fontSize=12, alignment=TA_JUSTIFY)
#         text = "Congratulations! You are able to boost deployment performance up to " \
#             + "<u><b>" + str(round(self.list_performance_top3[0] / self.original_model_performance, 1)) + \
#             "</b></u>" + "<u><b>X</b></u> on your model with the most performant "
#         "Deep Learning acceleration optimization set."
#         paragraphReportHeader = Paragraph(text, psHeaderText)
#         self.elements.append(paragraphReportHeader)

#         spacer = Spacer(10, 15)
#         self.elements.append(spacer)

#         d = []
#         textData = ["Ranking", "Optimization Set", "Performance (sample/sec)"]

#         fontSize = 8
#         centered = ParagraphStyle(name="centered", alignment=TA_LEFT)
#         for text in textData:
#             ptext = "<font size='%s'><b>%s</b></font>" % (fontSize, text)
#             titlesTable = Paragraph(ptext, centered)
#             d.append(titlesTable)

#         data = [d]
#         formattedLineData = []

#         alignStyle = [ParagraphStyle(name="01", alignment=TA_LEFT),
#                       ParagraphStyle(name="02", alignment=TA_LEFT),
#                       ParagraphStyle(name="03", alignment=TA_LEFT), ]

#         list_ranking = [1, 2, 3]
#         list_ranking.append(self.original_model_ranking)
#         list_optimization_set = self.list_optimization_set_top3
#         list_optimization_set.append(['Default'])
#         list_performance = self.list_performance_top3
#         list_performance.append(self.original_model_performance)

#         list_optimization_set_display = []
#         for i in list_optimization_set:
#             i = str(i).replace("[", "").replace("]", "").replace(
#                 "'", "").replace('"', "").replace(",", " +")
#             i = i.replace("pytorch_", "")
#             i = i.replace(
#                 "inc_static_quant_fx", "Intel Neural Compressor Post-Training Static Quantization (FX)")
#             i = i.replace(
#                 "inc_static_quant_ipex", "Intel Neural Compressor Post-Training Static Quantization (IPEX)")
#             i = i.replace(
#                 "inc_dynamic_quant", "Intel Neural Compressor Post-Training Dynamic Quantization")
#             i = i.replace("channels_last", "Channels Last")
#             i = i.replace("ipex_fp32", "Intel Extension for PyTorch FP32")
#             i = i.replace("ipex_bf16", "Intel Extension for PyTorch BF16")
#             i = i.replace("ipex_bf16", "Intel Extension for PyTorch BF16")
#             i = i.replace("ipex_int8_static_quant",
#                           "Intel Extension for PyTorch INT8 Static Quantization")
#             i = i.replace("ipex_int8_dynamic_quant",
#                           "Intel Extension for PyTorch INT8 Dynamic Quantization")
#             i = i.replace("torchdynamo_", "TorchDynamo ")
#             i = i.replace("jit_script", "JIT Script")
#             i = i.replace("jit_trace", "JIT Trace")
#             i = i.replace("_ofi", " with JIT Optimize-for-Inference")
#             i = i.replace("mixed_precision_cpu",
#                           "Automatic Mixed Precision (CPU)")
#             i = i.replace("mixed_precision_cuda",
#                           "Automatic Mixed Precision (CUDA)")

#             list_optimization_set_display.append(i)
#         for i in range(len(list_ranking)):
#             lineData = [list_ranking[i], list_optimization_set_display[i], int(
#                 list_performance[i])]
#             columnNumber = 0
#             for item in lineData:
#                 ptext = "<font size='%s'>%s</font>" % (fontSize - 1, item)
#                 p = Paragraph(ptext, alignStyle[columnNumber])
#                 formattedLineData.append(p)
#                 columnNumber = columnNumber + 1
#             data.append(formattedLineData)
#             formattedLineData = []

#         # totalRow = ["The most performant optimization set is: " + list_optimization_set_display[0]]
#         # for item in totalRow:
#         #     ptext = "<font size='%s'>%s</font>" % (fontSize-1, item)
#         #     p = Paragraph(ptext, alignStyle[1])
#         #     formattedLineData.append(p)
#         # data.append(formattedLineData)

#         table = Table(data, colWidths=[50, 290, 120])
#         tStyle = TableStyle([
#             ('ALIGN', (0, 0), (0, -1), 'LEFT'),
#             ('VALIGN', (0, 0), (-1, -1), 'TOP'),
#             ("ALIGN", (1, 0), (1, -1), 'RIGHT'),
#             ('LINEABOVE', (0, 0), (-1, -1), 1, self.IntelBlueLight),
#             ('BACKGROUND', (0, 0), (-1, 0), self.IntelBlueLight),
#             ('BACKGROUND', (0, 5), (0, 0), Color(
#                 (225 / 255), (225 / 255), (225 / 255), 1)),
#             ('BACKGROUND', (3, 5), (0, 3), Color(
#                 (240 / 255), (240 / 255), (240 / 255), 1)),
#             ('LINEBELOW', (0, 0), (-1, -1), 1, self.IntelBlueLight),
#         ])
#         table.setStyle(tStyle)
#         self.elements.append(table)

#         spacer = Spacer(10, 2)
#         self.elements.append(spacer)
#         psHeaderText = ParagraphStyle('Hed0', fontSize=7, alignment=TA_JUSTIFY)
#         from neural_coder.utils.cpu_info import get_num_cpu_cores
#         text = "*(1) All optimization sets are measured with default configuration " \
#                "(single instance on single socket), among which the top 3 performant ones are displayed. " \
#                "(2) This report evaluates performance only (accuracy under development)."
#         paragraphReportHeader = Paragraph(text, psHeaderText)
#         self.elements.append(paragraphReportHeader)

#         spacer = Spacer(10, 10)
#         self.elements.append(spacer)

#         # Configuration Table

#         psHeaderText = ParagraphStyle(
#             'Hed0', fontSize=16, alignment=TA_LEFT, borderWidth=3, textColor=self.IntelBlueDark)
#         text = '<b>Best Optimization (Sweeping Configs)</b>'
#         paragraphReportHeader = Paragraph(text, psHeaderText)
#         self.elements.append(paragraphReportHeader)

#         spacer = Spacer(5, 11)
#         self.elements.append(spacer)

#         psHeaderText = ParagraphStyle(
#             'Hed0', fontSize=12, alignment=TA_JUSTIFY)
#         text = "For the most performant optimization set, " \
#             "you can further boost your deployment performance to up to " \
#             + "<u><b>" + \
#             str(round(self.list_config_best_performance[0] / self.original_model_performance, 1)) + "</b></u>" + \
#             "<u><b>X</b></u> if using the most performant deployment configuration according to our sweeping result."
#         paragraphReportHeader = Paragraph(text, psHeaderText)
#         self.elements.append(paragraphReportHeader)

#         spacer = Spacer(10, 15)
#         self.elements.append(spacer)

#         d = []
#         textData = ["Category", "Num Intances",
#                     "Num Cores Per Instance", "BS", "Performance (sample/sec)"]

#         fontSize = 8
#         centered = ParagraphStyle(name="centered", alignment=TA_LEFT)
#         for text in textData:
#             ptext = "<font size='%s'><b>%s</b></font>" % (fontSize, text)
#             titlesTable = Paragraph(ptext, centered)
#             d.append(titlesTable)

#         data = [d]
#         formattedLineData = []

#         alignStyle = [ParagraphStyle(name="01", alignment=TA_LEFT),
#                       ParagraphStyle(name="02", alignment=TA_LEFT),
#                       ParagraphStyle(name="03", alignment=TA_LEFT),
#                       ParagraphStyle(name="04", alignment=TA_LEFT),
#                       ParagraphStyle(name="05", alignment=TA_LEFT)]

#         list_category = ["Throughput",
#                          "Throughput based on P50-Latency",
#                          "Throughput based on P90-Latency",
#                          "Throughput based on P99-Latency"]

#         for i in range(len(list_category)):
#             lineData = [list_category[i],
#                         self.list_config_best_ncpi[i],
#                         self.list_config_best_nins[i],
#                         self.list_config_best_bs[i],
#                         int(self.list_config_best_performance[i])]
#             columnNumber = 0
#             for item in lineData:
#                 ptext = "<font size='%s'>%s</font>" % (fontSize - 1, item)
#                 p = Paragraph(ptext, alignStyle[columnNumber])
#                 formattedLineData.append(p)
#                 columnNumber = columnNumber + 1
#             data.append(formattedLineData)
#             formattedLineData = []

#         table = Table(data, colWidths=[140, 65, 105, 30, 120])
#         tStyle = TableStyle([
#             ('ALIGN', (0, 0), (0, -1), 'LEFT'),
#             ('VALIGN', (0, 0), (-1, -1), 'TOP'),
#             ("ALIGN", (1, 0), (1, -1), 'RIGHT'),
#             ('LINEABOVE', (0, 0), (-1, -1), 1, self.IntelBlueLight),
#             ('BACKGROUND', (0, 0), (-1, 0), self.IntelBlueLight),
#             ('BACKGROUND', (0, 5), (0, 0), Color(
#                 (235 / 255), (235 / 255), (235 / 255), 1)),
#             ('LINEBELOW', (0, 0), (-1, -1), 1, self.IntelBlueLight),
#         ])
#         table.setStyle(tStyle)
#         self.elements.append(table)

#         spacer = Spacer(10, 2)
#         self.elements.append(spacer)
#         psHeaderText = ParagraphStyle('Hed0', fontSize=7, alignment=TA_JUSTIFY)
#         text = "*Measured on the most performant optimization set (Ranking 1 in above table) " \
#                "by sweeping configurations among batch size, number of instances, and number of cores per instance."
#         paragraphReportHeader = Paragraph(text, psHeaderText)
#         self.elements.append(paragraphReportHeader)

#         spacer = Spacer(10, 10)
#         self.elements.append(spacer)

#         # TCO

#         psHeaderText = ParagraphStyle(
#             'Hed0', fontSize=16, alignment=TA_LEFT, borderWidth=3, textColor=self.IntelBlueDark)
#         text = '<b>Cost Saving</b>'
#         paragraphReportHeader = Paragraph(text, psHeaderText)
#         self.elements.append(paragraphReportHeader)

#         # spacer = Spacer(5, 11)
#         # self.elements.append(spacer)

#         # psHeaderText = ParagraphStyle('Hed0', fontSize=12, alignment=TA_LEFT)
#         # text = "Samples per Dollar:"
#         # paragraphReportHeader = Paragraph(text, psHeaderText)
#         # self.elements.append(paragraphReportHeader)

#         spacer = Spacer(5, 11)
#         self.elements.append(spacer)

#         from reportlab.graphics.shapes import Drawing
#         from reportlab.graphics.charts.barcharts import HorizontalBarChart
#         from reportlab.graphics.charts.textlabels import Label
#         drawing = Drawing(0, 80)
#         TCO_raw = int(self.original_model_performance /
#                       (self.TCO_unit_pricing / 3600))
#         TCO_accelerated = int(
#             self.list_config_best_performance[0] / (self.TCO_unit_pricing / 3600))
#         data = [
#             (TCO_raw, TCO_accelerated)
#         ]
#         bc = HorizontalBarChart()
#         bc.valueAxis.labels.fontName = 'Helvetica'
#         bc.barLabels.fontName = 'Helvetica'
#         bc.categoryAxis.labels.fontName = 'Helvetica'
#         bc.x = 180
#         bc.y = 10
#         bc.height = 65
#         bc.width = 250
#         bc.data = data
#         bc.strokeColor = Color((235 / 255), (235 / 255), (235 / 255), 1)
#         bc.bars[(0, 0)].fillColor = self.IntelBlueDark
#         bc.bars[(0, 1)].fillColor = self.IntelBlueLight
#         bc.valueAxis.valueMin = 0
#         import math
#         bc.valueAxis.valueMax = max(TCO_accelerated, TCO_raw)
#         # bc.valueAxis.valueStep = math.ceil(TCO_raw)/10
#         bc.categoryAxis.labels.boxAnchor = 'ne'
#         bc.categoryAxis.labels.dx = -10
#         bc.categoryAxis.labels.dy = 4
#         bc.categoryAxis.labels.angle = 0
#         bc.categoryAxis.categoryNames = ["Default: " + str(format(TCO_raw, ',')) + " (sample/$)",
#                                          "Optimized: " + str(format(TCO_accelerated, ',')) + " (sample/$)"]
#         drawing.add(bc)
        
#         # add label
#         lab = Label()
#         lab.setOrigin(0, 0)
#         lab.boxAnchor = 'ne'
#         lab.angle = 0
#         lab.dx = 420
#         lab.dy = 35
#         lab.fontName = 'Helvetica-Bold'
#         lab.fontSize = 13
#         lab.setText(str(round(self.list_config_best_performance[0] / self.original_model_performance, 1)) + 'X')
#         drawing.add(lab)

#         self.elements.append(drawing)

#         spacer = Spacer(10, 9)
#         self.elements.append(spacer)
#         psHeaderText = ParagraphStyle('Hed0', fontSize=8, alignment=TA_JUSTIFY)
#         text = "*Sample/$ is calculated based on " + self.cloud_vendor + \
#             " " + self.cloud_instance_type + " instance and on-demand price."
#         paragraphReportHeader = Paragraph(text, psHeaderText)
#         self.elements.append(paragraphReportHeader)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re
import yaml

try:
    from ...conf.dotdict import DotDict
except:
    from .dot_dict import DotDict  ##TODO
from .logger import logger


def check_config(prune_config):
    assert prune_config['start_step'] >= 0, "start_step should be greater than 0"
    assert prune_config['end_step'] >= -1, "end_step should be greater than 0"
    assert prune_config['end_step'] >= prune_config['start_step'], \
        "end_step should be greater than start_step"
    assert prune_config['target_sparsity'] >= 0 and prune_config['target_sparsity'] < 1.0, \
        "begin_pruning_step should be in range [0,1)"
    assert prune_config['update_frequency_on_step'] > 0, "update_frequency_on_step should be greater than 0"
    assert prune_config['max_sparsity_ratio_per_layer'] >= 0 and prune_config['max_sparsity_ratio_per_layer'] < 1, \
        "update_frequency_on_step should be greater than 0"
    assert prune_config['prune_domain'] == "global" or prune_config['prune_domain'] == "local", \
        "only support 'global' and 'local' prune domain"
    if "x" in prune_config["pattern"]:
        pattern = prune_config["pattern"].split('_')[-1].split('x')
        N = int(pattern[0])
        M = int(pattern[1])
        assert N > 0, "N should be greater than 0"
        assert M > 0, "M should be greater than 0"
    if ":" in prune_config["pattern"]:
        pattern = prune_config["pattern"].split('_')[-1].split(':')
        N = int(pattern[0])
        M = int(pattern[1])
        assert N > 0, "N should be greater than 0"
        assert M > N, "M should be greater than N"
        max_ratio = float(N) / M
        assert prune_config['target_sparsity'] <= max_ratio, \
            "in N:M pattern, the max sparsity is N/M={}".format(max_ratio)
        prune_config['max_sparsity_ratio_per_layer'] = min(max_ratio, prune_config['max_sparsity_ratio_per_layer'])

def reset_non_value_to_default(obj, key, default):
     if isinstance(obj, dict):
        if (not key in obj.keys()) or obj[key] == None:
            return default
        else:
            return obj[key]
     else:
         if not hasattr(obj, key) or getattr(obj, key) == None:
             return default
         else:
             return getattr(obj, key)

def process_and_check_config(val):
    val = val["pruning"]['approach']['weight_compression_pytorch']
    start_step = reset_non_value_to_default(val, "start_step", 0)
    end_step = reset_non_value_to_default(val, "end_step", 0)
    not_to_prune_names = reset_non_value_to_default(val, "not_to_prune_names", [])
    prune_layer_type = reset_non_value_to_default(val, "prune_layer_type", ['Conv2d', 'Linear'])
    target_sparsity = reset_non_value_to_default(val, "target_sparsity", 0.0)  ## be care of this val
    update_frequency_on_step = int(reset_non_value_to_default(val, "update_frequency_on_step", 1))
    prune_domain = reset_non_value_to_default(val, "prune_domain", "global")
    prune_type = reset_non_value_to_default(val, "prune_type", "snip_momentum")
    sparsity_decay_type = reset_non_value_to_default(val, "sparsity_decay_type", "exp")
    max_sparsity_ratio_per_layer = reset_non_value_to_default(val, "max_sparsity_ratio_per_layer", 0.98)
    names = reset_non_value_to_default(val, "names", [])
    exclude_names = reset_non_value_to_default(val, "exclude_names", [])
    pattern = reset_non_value_to_default(val, "pattern", "tile_pattern_4x1")

    pruners_info = []
    for info in val['pruners']:
        pruner = {}
        pruner['start_step'] = reset_non_value_to_default(info, 'start_step', start_step)
        pruner['end_step'] = reset_non_value_to_default(info, 'end_step', end_step)
        pruner['not_to_prune_names'] = reset_non_value_to_default(info, 'not_to_prune_names', not_to_prune_names)
        pruner['prune_layer_type'] = reset_non_value_to_default(info, 'prune_layer_type', prune_layer_type)
        pruner['target_sparsity'] = reset_non_value_to_default(info, 'target_sparsity', target_sparsity)
        pruner['update_frequency_on_step'] = reset_non_value_to_default(info, 'update_frequency_on_step', \
                                                                 update_frequency_on_step)
        pruner['prune_domain'] = reset_non_value_to_default(info, 'prune_domain', prune_domain)
        pruner['prune_type'] = reset_non_value_to_default(info, 'prune_type', prune_type)
        pruner['sparsity_decay_type'] = reset_non_value_to_default(info, 'sparsity_decay_type', sparsity_decay_type)
        pruner['max_sparsity_ratio_per_layer'] = reset_non_value_to_default(info, 'max_sparsity_ratio_per_layer', \
                                                                 max_sparsity_ratio_per_layer)
        pruner['names'] = reset_non_value_to_default(info, 'names', names)
        pruner['exclude_names'] = reset_non_value_to_default(info, 'exclude_names',
                                                  exclude_names)
        pruner['pattern'] = reset_non_value_to_default(info, 'pattern',
                                            pattern)                                                  

        check_config(pruner)
        pruner_info = DotDict(pruner)
        pruners_info.append(pruner_info)
    return pruners_info


def process_config(config):
    if isinstance(config, str):
        try:
            with open(config, 'r') as f:
                content = f.read()
                try:
                    from .schema_check import schema
                except ImportError:
                    from ...conf.config import schema
                val = yaml.safe_load(content)
                schema.validate(val)
        except FileNotFoundError as f:
            logger.error("{}.".format(f))
            raise RuntimeError(
                "The yaml file is not exist. Please check the file name or path."
            )
        except Exception as e:
            logger.error("{}.".format(e))
            raise RuntimeError(
                "The yaml file format is not correct. Please refer to document."
            )

    elif isinstance(config, DotDict):
        val = config
    else:
        assert False, f"not supported type {config}"

    return process_and_check_config(val)


def parse_to_prune(model, config):
    """keep target pruned layers"""
    modules = {}
    if config["names"] == None or config["names"] == []:
        config["names"] = [".*"]
    for raw in config["names"]:
        try:
            pattern = re.compile(raw)
        except:
            assert False, f"regular expression match does not support {raw}"
        for name, module in filter(lambda t: pattern.search(t[0]), model.named_modules()):
            if type(module).__name__ in config["prune_layer_type"]:
                modules[name] = module
    return modules


def parse_not_to_prune(modules, config):
    """drop non pruned layers"""
    not_to_prune = config["exclude_names"]
    not_to_prune.extend(config["not_to_prune_names"])

    patterns = [re.compile(s) for s in not_to_prune]
    if len(patterns) <= 0:
        return modules
    new_module = {}
    for name in modules.keys():
        if any([p.search(name) for p in patterns]):
            continue
        new_module[name] = modules[name]
    return new_module



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import collections
import numpy as np
from abc import abstractmethod
from .sampler import IterableSampler, SequentialSampler, BatchSampler
from .fetcher import FETCHERS
from .base_dataloader import BaseDataLoader

def default_collate(batch):
    """Puts each data field into a pd frame with outer dimension batch size"""
    elem = batch[0]
    if isinstance(elem, collections.abc.Mapping):
        return {key: default_collate([d[key] for d in batch]) for key in elem}
    elif isinstance(elem, collections.abc.Sequence):
        batch = zip(*batch)
        return [default_collate(samples) for samples in batch]
    elif isinstance(elem, np.ndarray):
        try:
            return np.stack(batch)
        except:
            return batch
    else:
        return batch

class DefaultDataLoader(BaseDataLoader):
    """DefaultDataLoader

    """

    def __init__(self, dataset, batch_size=1, last_batch='rollover', collate_fn=None,
                 sampler=None, batch_sampler=None, num_workers=0, pin_memory=False,
                 shuffle=False, distributed=False):
        self.dataset = dataset
        self.last_batch = last_batch
        self.sampler = sampler
        self.batch_sampler = batch_sampler
        self.num_workers = num_workers
        self.pin_memory = pin_memory
        self.collate_fn = collate_fn
        self._batch_size = batch_size
        self.shuffle = shuffle
        self.distributed = distributed
        if self.collate_fn == None:
            self.collate_fn = default_collate

    def batch(self, batch_size, last_batch='rollover'):
        self._batch_size = batch_size
        self.last_batch = last_batch

    @property
    def dataloader(self):
        return self

    def __iter__(self):
        return self._generate_dataloader(
            self.dataset,
            batch_size=self.batch_size,
            last_batch=self.last_batch,
            collate_fn=self.collate_fn,
            sampler=self.sampler,
            batch_sampler=self.batch_sampler,
            num_workers=self.num_workers,
            pin_memory=self.pin_memory,
            shuffle=self.shuffle,
            distributed=self.distributed)

    def _generate_dataloader(self, dataset, batch_size, last_batch, collate_fn, sampler,
                             batch_sampler, num_workers, pin_memory, shuffle, distributed):

        drop_last = False if last_batch == 'rollover' else True
        sampler = self._generate_sampler(dataset, distributed)
        self.batch_sampler = BatchSampler(sampler, batch_size, drop_last)
        self.fetcher = FETCHERS[self.dataset_type](dataset, collate_fn, drop_last, distributed)

        for batched_indices in self.batch_sampler:
            try:
                data = self.fetcher(batched_indices)
                yield data
            except StopIteration:
                return

    def _generate_sampler(self, dataset, distributed):
        if hasattr(dataset, "__getitem__"):
            self.dataset_type = 'index'
            return SequentialSampler(dataset, distributed)
        elif hasattr(dataset, "__iter__"):
            self.dataset_type = 'iter'
            return IterableSampler(dataset)
        else:
            raise ValueError("dataset type only support (index, iter)")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from neural_compressor.experimental.data.dataloaders import DATALOADERS

# THIS API IS TO BE DEPRECATED!
class DataLoader(object):
    """Entrance of all configured DataLoaders. Will dispatch the DataLoaders to framework
       specific one. Users will be not aware of the dispatching, and the Interface is unified.

    """

    def __new__(cls, framework, dataset, batch_size=1, collate_fn=None,
                 last_batch='rollover', sampler=None, batch_sampler=None,
                 num_workers=0, pin_memory=False, shuffle=False, distributed=False):

        assert framework in ('tensorflow', 'inteltensorflow', 'tensorflow_itex', \
                             'pytorch', 'pytorch_ipex', 'pytorch_fx', 'onnxrt_qdqops', \
                             'onnxrt_qlinearops', 'onnxrt_integerops', 'mxnet'), \
                             "framework support tensorflow pytorch mxnet onnxruntime"
        return DATALOADERS[framework](dataset=dataset,
                                      batch_size=batch_size,
                                      last_batch=last_batch,
                                      collate_fn=collate_fn,
                                      sampler=sampler,
                                      batch_sampler=batch_sampler,
                                      num_workers=num_workers,
                                      pin_memory=pin_memory,
                                      shuffle=shuffle,
                                      distributed=distributed)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import abstractmethod
from enum import EnumMeta
import os
import math
import copy
import pickle
from collections import OrderedDict
from pathlib import Path
import yaml
import numpy as np
from ..objective import MultiObjective
from ..adaptor import FRAMEWORKS
from ..utils.utility import Statistics, dump_data_to_local
from ..utils.utility import fault_tolerant_file, equal_dicts, GLOBAL_STATE, MODE
from ..utils.create_obj_from_config import create_eval_func, create_train_func
from ..utils import logger
from ..utils import OPTIONS
from ..version import __version__
from ..conf.dotdict import DotDict, deep_get, deep_set
from ..algorithm import AlgorithmScheduler
from ..algorithm.fast_bias_correction import FastBiasCorrection

import copy
import numpy as np
from collections import OrderedDict
from ..utils import logger


from .st_utils.tuning_sampler import OpTypeWiseTuningSampler, FallbackTuningSampler
from .st_utils.tuning_space import TuningItem, TuningSpace
from .st_utils.tuning_structs import OpTuningConfig


STRATEGIES = {}


def strategy_registry(cls):
    """The class decorator used to register all TuneStrategy subclasses.

    Args:
        cls (class): The class of register.

    Returns:
        cls: The class of register.
    """
    assert cls.__name__.endswith(
        'TuneStrategy'
    ), "The name of subclass of TuneStrategy should end with \'TuneStrategy\' substring."
    if cls.__name__[:-len('TuneStrategy')].lower() in STRATEGIES:
        raise ValueError('Cannot have two strategies with the same name')
    STRATEGIES[cls.__name__[:-len('TuneStrategy')].lower()] = cls
    return cls

@strategy_registry
class TuneStrategy(object):
    """The base class of tuning strategy.

    Args:
        model (object):                        The FP32 model specified for low precision tuning.
        conf (Conf):                           The Conf class instance initialized from user yaml
                                               config file.
        q_dataloader (generator):              Data loader for calibration, mandatory for
                                               post-training quantization.
                                               It is iterable and should yield a tuple (input,
                                               label) for calibration dataset containing label,
                                               or yield (input, _) for label-free calibration
                                               dataset. The input could be a object, list, tuple or
                                               dict, depending on user implementation, as well as
                                               it can be taken as model input.
        q_func (function, optional):           Reserved for future use.
        eval_dataloader (generator, optional): Data loader for evaluation. It is iterable
                                               and should yield a tuple of (input, label).
                                               The input could be a object, list, tuple or dict,
                                               depending on user implementation, as well as it can
                                               be taken as model input. The label should be able
                                               to take as input of supported metrics. If this
                                               parameter is not None, user needs to specify
                                               pre-defined evaluation metrics through configuration
                                               file and should set "eval_func" parameter as None.
                                               Tuner will combine model, eval_dataloader and
                                               pre-defined metrics to run evaluation process.
        eval_func (function, optional):        The evaluation function provided by user.
                                               This function takes model as parameter, and
                                               evaluation dataset and metrics should be
                                               encapsulated in this function implementation and
                                               outputs a higher-is-better accuracy scalar value.

                                               The pseudo code should be something like:

                                               def eval_func(model):
                                                    input, label = dataloader()
                                                    output = model(input)
                                                    accuracy = metric(output, label)
                                                    return accuracy
        resume(dict, optional):                The dict containing resume information.
                                               Defaults to None.
    """


    def __init__(self, model, conf, q_dataloader=None, q_func=None,
                 eval_dataloader=None, eval_func=None, resume=None, q_hooks=None):
        self.model = model
        self.cfg = conf.usr_cfg
        self.history_path = self.create_path(self.cfg.tuning.workspace.path, './history.snapshot')
        self.deploy_path = self.create_path(self.cfg.tuning.workspace.path, 'deploy.yaml')
        logger.debug("Dump user yaml configuration:")
        logger.debug(self.cfg)

        self.eval_dataloader = eval_dataloader
        self.calib_dataloader = q_dataloader
        self.q_func = q_func
        self.q_hooks = q_hooks
        self.eval_func = eval_func
        GLOBAL_STATE.STATE = MODE.QUANTIZATION
        framework, framework_specific_info = self.set_framework_info(q_dataloader)
        self.adaptor = FRAMEWORKS[framework](framework_specific_info)
        self.framework = framework

        self.set_q_func()
        self.set_objectives()
        self.tune_data = {}
        self.tune_result_record = []
        self.tuning_history = []
        # The tuning history ever made, structured like below:
        # [
        #   {
        #     'version': __version__,
        #     'cfg': cfg1,
        #     'framework': tensorflow
        #     'baseline': baseline1,
        #     'last_tune_result': last_tune_result1,
        #     'best_tune_result': best_tune_result1,
        #     'history': [
        #                  # tuning history under same yaml config
        #                  {'tune_cfg': tune_cfg1, 'tune_result': \
        #                               tune_result1, 'q_config': q_config1, ...},

        #                   ...,
        #                ],
        #     # new fields added by subclass for resuming
        #     ...,
        #   },
        #   # tuning history under different yaml configs
        #   ...,
        # ]

        self.baseline = None
        self.last_tune_result = None
        self.last_qmodel = None
        self.best_tune_result = None
        self.best_qmodel = None
        self.cur_best_acc = self.initial_best_acc() # track the current best accuracy
        self.cur_best_tuning_cfg = {} # track tuning cfg with the current best accuracy

        self.capability = self.adaptor.query_fw_capability(model)
        logger.debug(self.capability)
        self.set_tuning_space(conf)

        self.algo = AlgorithmScheduler(self.cfg.quantization.recipes)
        self.algo.dataloader = self.calib_dataloader  # reuse the calibration iteration
        self.algo.origin_model = self.model
        self.algo.adaptor = self.adaptor
        
        if resume is not None: self.setup_resume(resume)


    @abstractmethod
    def next_tune_cfg(self):
        """The generator of yielding next tuning config to traverse by concrete strategies
           according to last tuning result.

        Yields:
            tune_config (dict): It's a dict containing the tuning configuration to run.
        """
        raise NotImplementedError
    

    def traverse(self):
        """The main traverse logic, which could be override by some concrete strategy which needs
           more hooks.
        """
        if not (self.cfg.evaluation and self.cfg.evaluation.accuracy and \
            (self.cfg.evaluation.accuracy.metric or self.cfg.evaluation.accuracy.multi_metrics)) \
            and self.eval_func is None:
            logger.info("Neither evaluation function nor metric is defined." \
                        " Generate a quantized model with default quantization configuration.")
            self.cfg.tuning.exit_policy.performance_only = True
            logger.info("Force setting 'tuning.exit_policy.performance_only = True'.")
            logger.info("Generate a fake evaluation function.")
            self.eval_func = self._fake_eval_func

        # get fp32 model baseline
        if self.baseline is None:
            logger.info("Get FP32 model baseline.")
            self._fp32_model = self.model
            self.baseline = self._evaluate(self.model)
            # record the FP32 baseline
            self._add_tuning_history()
        self.show_baseline_info()

        trials_count = 0
        for op_tuning_cfg in self.next_tune_cfg():
            tune_cfg = self._tune_cfg_converter(op_tuning_cfg)
            trials_count += 1
            tuning_history = self._find_tuning_history(tune_cfg)
            if tuning_history and trials_count < self.cfg.tuning.exit_policy.max_trials:
                self.last_tune_result = tuning_history['last_tune_result']
                self.best_tune_result = tuning_history['best_tune_result']
                logger.warn("Find evaluated tuning config, skip.")
                continue
            logger.debug("Dump current tuning configuration:")
            logger.debug(tune_cfg)
            self.q_model = self.adaptor.quantize(
                copy.deepcopy(tune_cfg), self.model, self.calib_dataloader, self.q_func)
            self.algo.calib_iter = tune_cfg['calib_iteration']
            self.algo.q_model = self.q_model
            # TODO align the api to let strategy has access to pre_optimized model
            assert self.adaptor.pre_optimized_model
            self.algo.origin_model = self.adaptor.pre_optimized_model
            if self.cfg.quantization.recipes.fast_bias_correction:
                self.algo.algorithms[0].quantization_cfg = tune_cfg
            self.last_qmodel = self.algo()
            assert self.last_qmodel
            self.last_tune_result = self._evaluate(self.last_qmodel)
            self.cur_best_acc, self.cur_best_tuning_cfg = self.update_best_op_tuning_cfg(op_tuning_cfg)
            need_stop = self.stop(self.cfg.tuning.exit_policy.timeout, trials_count)

            # record the tuning history
            saved_tune_cfg = copy.deepcopy(tune_cfg)
            saved_last_tune_result = copy.deepcopy(self.last_tune_result)
            self._add_tuning_history(saved_tune_cfg,
                                    saved_last_tune_result,
                                    q_config=self.q_model.q_config)
            self.tune_result_record.append(copy.deepcopy(self.last_tune_result))
            self.tune_cfg = tune_cfg
            if need_stop:
                if self.cfg.tuning.diagnosis and self.cfg.tuning.diagnosis.diagnosis_after_tuning:
                    logger.debug(f'*** Start to do diagnosis (inspect tensor).')
                    self._diagnosis()
                if self.use_multi_objective and len(self.tune_result_record) > 1 and \
                    self.best_tune_result is not None:
                    best_trail, best_result = self.objectives.best_result(self.tune_result_record,
                                                                          copy.deepcopy(self.baseline))
                    if best_result != self.best_tune_result:
                        from neural_compressor.utils.utility import recover
                        self.best_qmodel = recover(self.model.model, 
                            os.path.join(self.cfg.tuning.workspace.path, 'history.snapshot'),
                            best_trail)
                        self.best_tune_result = best_result
                break

    def initial_tuning_cfg(self):
        if self.cfg.quantization.approach == 'post_training_auto_quant':
            query_order = ['static', 'dynamic', 'bf16', 'fp16', 'fp32']
        elif self.cfg.quantization.approach == 'post_training_dynamic_quant':
            query_order = ['dynamic', 'bf16', 'fp16', 'fp32']
        elif self.cfg.quantization.approach == 'post_training_static_quant':
            query_order = ['static', 'bf16', 'fp16', 'fp32']
        elif self.cfg.quantization.approach == 'quant_aware_training':
            query_order = ['static', 'dynamic', 'bf16', 'fp16', 'fp32']

        quant_mode_wise_items = OrderedDict()
        pre_items = set()
        for quant_mode in query_order:
            items = self.tuning_space.query_items_by_quant_mode(quant_mode)
            filtered_items = [item for item in items if item not in pre_items]
            pre_items = pre_items.union(set(items))
            quant_mode_wise_items[quant_mode] = filtered_items

        def initial_op_quant_mode(items_lst, target_quant_mode, op_item_dtype_dict):
            for item in items_lst:
                op_item_dtype_dict[item.name] = target_quant_mode

        op_item_dtype_dict = OrderedDict()
        for quant_mode, quant_mode_items in quant_mode_wise_items.items():
            initial_op_quant_mode(quant_mode_items, quant_mode, op_item_dtype_dict)

        initial_op_tuning_cfg = {}
        for op_name_dtype, quant_mode in op_item_dtype_dict.items():
            initial_op_tuning_cfg[op_name_dtype] = OpTuningConfig(op_name_dtype[0], op_name_dtype[1], 
                                                                  quant_mode, self.tuning_space)
        return op_item_dtype_dict, quant_mode_wise_items, initial_op_tuning_cfg

    def show_baseline_info(self):
        if self.baseline:
            self.tune_data['baseline'] = self.baseline[0] if \
                isinstance(self.baseline[0], list) else [self.baseline[0]]
            for name, data in zip(self.metric_name, self.tune_data['baseline']):
                self.tune_data[name] = [data]
            if self.metric_weight:
                # baseline is weighted accuracy
                self.tune_data['Weighted accuracy'] = \
                    [np.mean(np.array(self.tune_data['baseline']) * self.metric_weight)]
                self.tune_data['baseline'] = self.tune_data['Weighted accuracy']
            baseline_msg = '[Accuracy:' + \
                ''.join([' {:.4f}'.format(i) for i in self.tune_data['baseline']]) + \
                ''.join([', {}: {:.4f}'.format(x,y) for x,y in zip( \
                self.objectives.representation, self.baseline[1]) if x != 'Accuracy']) + ']'
        else: # pragma: no cover
            if self.metric_weight:
                self.tune_data['Weighted accuracy'] = ['n/a']
            self.tune_data['baseline'] = ['n/a']

            for name, data in zip(self.metric_name, self.tune_data['baseline']):
                self.tune_data[name] = ['n/a']
            baseline_msg = 'n/a'
        logger.info("FP32 baseline is: {}".format(baseline_msg))

    def initial_best_acc(self):
        if len(self.metric_name) == 1 or self.metric_weight is not None:
            best_acc = float('-inf') if self.higher_is_better else float('inf')
        else:
            best_acc = [float('-inf') if higher_is_better else float('inf') for \
                        higher_is_better in self.metric_criterion]
        return best_acc

    def _tune_cfg_converter(self, op_tuning_cfg):
        """ convert op_tuning_cfg for adaptor

        Args:
            op_tuning_cfg (Dict): 
        """
        tune_cfg = {'op': OrderedDict()}
        for op_name_type, op_config in op_tuning_cfg.items():
            if isinstance(op_config, OpTuningConfig):
                tune_cfg['op'][op_name_type] = op_config.get_state()
            else:
                tune_cfg[op_name_type] = op_config
        tune_cfg['calib_sampling_size'] = op_tuning_cfg['calib_sampling_size']
        if self.calib_dataloader:
            tune_cfg['calib_iteration'] =  math.ceil(int(tune_cfg['calib_sampling_size']) / \
                                                    self.calib_dataloader.batch_size) 
        else:
            tune_cfg['calib_iteration'] = 1
        tune_cfg['advance'] = self.cfg.quantization.advance
        return tune_cfg

    def set_tuning_space(self, conf):
        calib_sampling_size_lst = self.cfg.quantization.calibration.sampling_size
        calib_sampling_size_lst = [int(calib_sampling_size) for calib_sampling_size in calib_sampling_size_lst]
        if self.calib_dataloader:
            self.calib_iter = [math.ceil(int(x) / self.calib_dataloader.batch_size) \
                               for x in calib_sampling_size_lst]
        else:
            self.calib_iter = 1
        # create tuning space
        adaptor_cap = {
            'calib': {'calib_sampling_size': calib_sampling_size_lst},
            'op': self.capability['opwise']
        }
        self.tuning_space = TuningSpace(adaptor_cap, conf=conf, framework=self.framework)
        if logger.level < 20:
            self.tuning_space.root_item.show_details()

    def setup_resume(self, resume):
        self.__dict__.update(resume)
        for history in self.tuning_history:
            if self._same_yaml(history['cfg'], self.cfg):
                self.__dict__.update({k: v for k, v in history.items() \
                                        if k not in ['version', 'history']})
                logger.info("Start to resume tuning process.")
                # resume the best tuning model if needed
                try:
                    index = history['id'] - 1
                    resume_tuning_cfg = history['history'][index]['tune_cfg']
                    self.best_qmodel = self.adaptor.quantize(resume_tuning_cfg,
                                                                self.model, 
                                                                self.calib_dataloader, 
                                                                self.q_func)
                except:
                    logger.debug("Can not resume the best quantize model from history.")
                    
                break

    def set_q_func(self):
        if self.q_func == None and self.cfg.quantization.approach == 'quant_aware_training':
            train_cfg = self.cfg.quantization.train
            assert train_cfg, "train field of quantization section in yaml file must " \
                              "be configured for quantization aware training if q_func is NOT set."
            assert self.calib_dataloader, "dataloader field of train field of quantization " \
                                          "section in yaml file must be configured."
            self.q_func = create_train_func(self.framework, self.calib_dataloader, \
                                            self.adaptor, train_cfg, hooks=self.q_hooks)

    def create_path(self, custom_path, filename):
        new_path = os.path.join(os.path.abspath(os.path.expanduser(custom_path)),filename)
        path = Path(os.path.dirname(new_path))
        path.mkdir(exist_ok=True, parents=True)
        return new_path

    def set_framework_info(self,q_dataloader):
        framework_specific_info = {'device': self.cfg.device,
                                   'approach': self.cfg.quantization.approach,
                                   'random_seed': self.cfg.tuning.random_seed}
        framework = self.cfg.model.framework.lower()

        self.mixed_precision_mode = bool('mixed_precision' in self.cfg) or \
            bool('graph_optimization' in self.cfg)

        if 'tensorflow' in framework:
            framework_specific_info.update(
                {"inputs": self.cfg.model.inputs,
                 "outputs": self.cfg.model.outputs,
                 'workspace_path': self.cfg.tuning.workspace.path,
                 'recipes': self.cfg.quantization.recipes,
                 'performance_only': self.cfg.tuning.exit_policy.performance_only})
        if framework == 'mxnet':
            framework_specific_info.update({"q_dataloader": q_dataloader})
        if 'onnxrt' in framework.lower():
            if self.mixed_precision_mode:
                framework_specific_info.update({"backend": "integerops"})
                framework_specific_info.update({"approach": "post_training_dynamic_quant"})
            else:
                framework_specific_info.update({"backend": framework.lower().split('_')[-1]})
            framework_specific_info.update({"deploy_path": os.path.dirname(self.deploy_path)})
            framework_specific_info.update({'workspace_path': self.cfg.tuning.workspace.path})
            framework_specific_info.update({'recipes': self.cfg.quantization.recipes})
            framework_specific_info.update(
                                {'graph_optimization': OPTIONS[framework].graph_optimization})
            framework_specific_info.update({'reduce_range': self.cfg.reduce_range})
        if framework == 'pytorch_ipex' or framework == 'pytorch' or framework == 'pytorch_fx':
            if self.mixed_precision_mode:
                framework_specific_info.update({"approach": "post_training_dynamic_quant"})
            framework_specific_info.update({"q_dataloader": q_dataloader})
            framework_specific_info.update({"use_bf16": self.cfg.use_bf16 \
                            if self.cfg.use_bf16 is not None else True})
            framework_specific_info.update(
                {"workspace_path": os.path.dirname(self.deploy_path)})
            if self.cfg['quantization']['op_wise'] is not None \
               and 'default_qconfig' in self.cfg['quantization']['op_wise']:
                framework_specific_info.update(
                    {"default_qconfig": self.cfg['quantization']['op_wise']['default_qconfig']})
        return framework, framework_specific_info

    def set_objectives(self):
        self.higher_is_better = bool(self.cfg.tuning.accuracy_criterion.higher_is_better)
        self.use_multi_objective = deep_get(self.cfg, 'tuning.multi_objectives') and \
            len(self.cfg.tuning.multi_objectives.objective) > 1
        objectives = [i.lower() for i in self.cfg.tuning.multi_objectives.objective] if \
            self.use_multi_objective else [self.cfg.tuning.objective.lower()]
        self.metric_weight = deep_get(self.cfg, 'evaluation.accuracy.multi_metrics.weight')
        self.metric_name = ['Accuracy'] if \
            not deep_get(self.cfg, 'evaluation.accuracy.multi_metrics') else \
            self.cfg.evaluation.accuracy.multi_metrics.keys()-{'weight','higher_is_better'}
        if len(self.metric_name) == 1:
            self.metric_criterion = [self.higher_is_better]
        elif not deep_get(self.cfg, 'evaluation.accuracy.multi_metrics.higher_is_better'):
            # default is True
            self.metric_criterion = [True] * len(self.metric_name)
        else:
            self.metric_criterion = \
                deep_get(self.cfg, 'evaluation.accuracy.multi_metrics.higher_is_better')

        self.objectives = MultiObjective(objectives, 
                             self.cfg.tuning.accuracy_criterion,
                             self.metric_criterion,
                             self.metric_weight,
                             deep_get(self.cfg, 'tuning.multi_objectives.higher_is_better'),
                             deep_get(self.cfg, 'tuning.multi_objectives.weight'))

    def _same_yaml(self, src_yaml, dst_yaml):
        """Check whether two yamls are same, excluding those keys which does not really
           impact tuning result, such as tensorboard, workspace, resume options under tuning
           section of yaml.
        """
        if equal_dicts(src_yaml, dst_yaml, ignore_keys=['tuning']) and \
           equal_dicts(src_yaml.tuning, src_yaml.tuning, compare_keys=['objective',
                                                                       'accuracy_criterion',
                                                                       'random_seed',
                                                                       'exit_policy']):
            return True

        return False

    def update_best_op_tuning_cfg(self, op_tuning_cfg):
        # track and update the best tuning cfg/acc
        acc, _ = self.last_tune_result
        if self.cur_best_tuning_cfg is None:
            self.cur_best_tuning_cfg = copy.deepcopy(op_tuning_cfg)
        if not isinstance(acc, list) and ((self.higher_is_better and acc >= self.cur_best_acc) \
            or (not self.higher_is_better and acc <= self.cur_best_acc)):
            self.cur_best_acc = acc
            self.cur_best_tuning_cfg = copy.deepcopy(op_tuning_cfg)
        elif len(self.metric_name) > 1 and self.metric_weight is not None:
            acc = np.mean(np.array(acc) * self.metric_weight)
            if (self.higher_is_better and acc >= self.cur_best_acc) or \
                (not self.higher_is_better and acc <= self.cur_best_acc):
                self.cur_best_acc = acc
                self.cur_best_tuning_cfg = copy.deepcopy(op_tuning_cfg)
        elif len(self.metric_name) > 1 and self.metric_weight is None:
            if all([acc_i >= best_i if higher_is_better else acc_i <= best_i for \
                acc_i, best_i, higher_is_better in \
                zip(acc, self.cur_best_acc, self.metric_criterion)]):
                self.cur_best_acc = acc
                self.cur_best_tuning_cfg = copy.deepcopy(op_tuning_cfg)
        logger.debug(f"Best acc is {self.cur_best_acc}.")
        return self.cur_best_acc, self.cur_best_tuning_cfg

    def deploy_config(self):
        acc_dataloader_cfg = deep_get(self.cfg, 'evaluation.accuracy.dataloader')
        perf_dataloader_cfg = deep_get(self.cfg, 'evaluation.performance.dataloader')
        # use acc dataloader if perf dataloader is not configured
        if perf_dataloader_cfg is None:
            perf_dataloader_cfg = acc_dataloader_cfg

        self.deploy_cfg = OrderedDict()
        # int8 dataloader graph transform
        if deep_get(perf_dataloader_cfg, 'transform.QuantizedInput') is not None \
          or deep_get(acc_dataloader_cfg, 'transform.QuantizedInput') is not None:
            self.best_qmodel, scale = self.adaptor.quantize_input(self.best_qmodel)
            deep_set(perf_dataloader_cfg, 'transform.QuantizedInput.dtype', 'int8')
            deep_set(perf_dataloader_cfg, 'transform.QuantizedInput.scale', scale)
            deep_set(acc_dataloader_cfg, 'transform.QuantizedInput.dtype', 'int8')
            deep_set(acc_dataloader_cfg, 'transform.QuantizedInput.scale', scale)

        self.deploy_cfg['model'] = self.cfg.model
        self.deploy_cfg['device'] = self.cfg.device
        if self.cfg.evaluation is not None:
            deep_set(self.cfg, 'evaluation.performance.dataloader',\
                perf_dataloader_cfg)
            deep_set(self.cfg, 'evaluation.accuracy.dataloader', \
                acc_dataloader_cfg)
            self.deploy_cfg['evaluation'] = self.cfg.evaluation

        def setup_yaml():
            represent_dict_order = lambda self, \
                data: self.represent_mapping('tag:yaml.org,2002:map', data.items())
            yaml.add_representer(OrderedDict, represent_dict_order)
            yaml.add_representer(DotDict, represent_dict_order)
        setup_yaml()
        with open(self.deploy_path, 'w+') as f:
            yaml.dump(self.deploy_cfg, f)
            logger.info("Save deploy yaml to {}".format(self.deploy_path))

    def _get_common_cfg(self, model_wise_cfg, op_wise_cfgs):
        """Get the common parts from the model_wise_cfg.
            This function is focused on composing the configuration that consists of
            model-wise field and op-wise unique field data.

        Args:
            model_wise_cfg ([DotDict]): The model-wise configuration.
            op_wise_cfgs ([List]): The list of each op's config in DotDict type.

        Returns:
            [DotDict]: The combined configration with the op-wise unique field.
        """
        model_wise_keys = model_wise_cfg.keys()

        result = op_wise_cfgs[0]
        for each_op_wise_cfg in op_wise_cfgs:
            tmp_cfg = {}
            for k in model_wise_keys:
                tmp_cfg[k] = each_op_wise_cfg[k]

            if model_wise_cfg == tmp_cfg:
                result = each_op_wise_cfg
                break

        return result

    @property
    def evaluation_result(self):
        return self._evaluate(self.model)

    def _evaluate(self, model):
        """The interface of evaluating model.

        Args:
            model (object): The model to be evaluated.

        Returns:
            Objective: The objective value evaluated
        """
        if self.eval_func:
            if self.cfg.tuning.tensorboard:
                # Pytorch can insert observer to model in this hook.
                # Tensorflow don't support this mode for now
                model = self.adaptor._pre_eval_hook(model)
            val = self.objectives.evaluate(
                self.eval_func, model if self.framework == "pytorch_ipex" else model.model
            )
            if self.cfg.tuning.tensorboard:
                # post_eval_hook to deal the tensor
                self.adaptor._post_eval_hook(model, accuracy=val[0])
        else:
            assert self.cfg.evaluation and self.cfg.evaluation.accuracy and \
                (self.cfg.evaluation.accuracy.metric or \
                self.cfg.evaluation.accuracy.multi_metrics), \
                "metric or multi_metrics field of accuracy field of evaluation" \
                " section should not be empty"

            postprocess_cfg = self.cfg.evaluation.accuracy.postprocess
            metric_cfg = self.cfg.evaluation.accuracy.metric if \
                self.cfg.evaluation.accuracy.metric else \
                self.cfg.evaluation.accuracy.multi_metrics
            eval_func = create_eval_func(self.framework,
                self.eval_dataloader,
                self.adaptor,
                metric_cfg,
                postprocess_cfg,
                self.cfg.evaluation.accuracy.iteration,
                tensorboard = self.cfg.tuning.tensorboard,
                fp32_baseline = self.baseline == None)

            if getattr(self.eval_dataloader, 'distributed', False):
                if 'tensorflow' in self.framework:
                    import horovod.tensorflow as hvd
                elif self.framework in ['pytorch_ipex','pytorch','pytorch_fx']:
                    import horovod.torch as hvd
                else:
                    raise NotImplementedError("Currently only TensorFlow and PyTorch "
                                              "support distributed inference in PTQ.")
                hvd.init()
                try:
                    len_dataloader = len(self.eval_dataloader)
                except:
                    logger.info("The length of the distributed dataloader is unknown."
                                "When the iteration of evaluation dataloader in each "
                                "process is inconsistent, an error may occur.")
                else:
                    list_len_dataloader = hvd.allgather_object(len_dataloader)
                    if hvd.rank() == 0:
                        for i in range(len(list_len_dataloader)-1):
                            if list_len_dataloader[i] != list_len_dataloader[i+1]:
                                raise AttributeError("The evaluation dataloader's iteration is"
                                                     "different between processes, please reset "
                                                     "dataloader's batch_size.")
            val = self.objectives.evaluate(eval_func, model) 
        if isinstance(val[0], list):
            assert all([np.isscalar(i) for i in val[0]]), \
                "The eval_func should return a scalar or list of scalar, " \
                "but not {}!".format(str([type(i) for i in val[0]]))
        else:
            assert np.isscalar(val[0]), \
                "The eval_func should return a scalar or list of scalar, " \
                "but not {}!".format(str(type(val[0])))
            
        return val

    def __getstate__(self):
        """Magic method for pickle saving.

        Returns:
            dict: Saved dict for resuming
        """
        return {'tuning_history': self.tuning_history}

    def __setstate__(self, d):
        """Magic method for pickle loading.

        Args:
            d (dict): The dict to load.
        """
        self.__dict__.update(d)

    def stop(self, timeout, trials_count):
        """Check if need to stop traversing the tuning space, either accuracy goal is met
           or timeout is reach.

        Returns:
            bool: True if need stop, otherwise False
        """
        need_stop = False
        if self.cfg.tuning.exit_policy.performance_only or \
            self.objectives.compare(self.best_tune_result, self.baseline):
            del self.best_tune_result
            del self.best_qmodel
            self.best_tune_result = self.last_tune_result
            self.best_qmodel = self.last_qmodel
        else:
            del self.last_qmodel

        if self.last_tune_result:
            last_tune = self.last_tune_result[0] if \
                isinstance(self.last_tune_result[0], list) else [self.last_tune_result[0]]

            for name, data in zip(self.metric_name, last_tune):
                if len(self.tune_data[name]) == 1:
                    self.tune_data[name].append(data)
                else:
                    self.tune_data[name][1] = data

            if self.metric_weight and len(last_tune) > 1:
                weighted_acc = np.mean(np.array(last_tune) * self.metric_weight)
                    
                if len(self.tune_data['Weighted accuracy']) == 1:
                    self.tune_data['Weighted accuracy'].append(weighted_acc)
                else:
                    self.tune_data['Weighted accuracy'][1] = weighted_acc

                last_tune = [weighted_acc]

            last_tune_msg = '[Accuracy (int8|fp32):' + \
                ''.join([' {:.4f}|{:.4f}'.format(last, base) for last, base in \
                zip(last_tune, self.tune_data['baseline'])]) + \
                ''.join([', {} (int8|fp32): {:.4f}|{:.4f}'.format( \
                x, y, z) for x, y, z in zip( \
                self.objectives.representation, self.last_tune_result[1], self.baseline[1]) \
                if x != 'Accuracy']) + ']'
        else: # pragma: no cover
            last_tune_msg = 'n/a'
            for name in self.tune_data.keys() - {'baseline'}:
                if len(self.tune_data[name]) == 1:
                    self.tune_data[name].append('n/a')
                else:
                    self.tune_data[name][1] = 'n/a'

        if self.best_tune_result:
            best_tune = self.best_tune_result[0] if isinstance(self.best_tune_result[0], list) \
                        else [self.best_tune_result[0]]
 
            for name, data in zip(self.metric_name, best_tune):
                if len(self.tune_data[name]) == 2:
                     self.tune_data[name].append(data)
                else:
                    self.tune_data[name][2] = data

            if self.metric_weight and len(best_tune) > 1:
                weighted_acc = np.mean(np.array(best_tune) * self.metric_weight)

                if len(self.tune_data['Weighted accuracy']) == 2:
                    self.tune_data['Weighted accuracy'].append(weighted_acc)
                else: # pragma: no cover
                    self.tune_data['Weighted accuracy'][2] = weighted_acc

                best_tune = [weighted_acc]

            best_tune_msg = '[Accuracy:' + ''.join([' {:.4f}'.format(best) \
                for best in best_tune]) + ''.join([', {}: {:.4f}'.format(x,y) \
                for x,y in zip(self.objectives.representation, \
                self.best_tune_result[1]) if x != 'Accuracy']) + ']'

        else:
            best_tune_msg = 'n/a'
            for name in self.tune_data.keys() - {'baseline'}:
                if len(self.tune_data[name]) == 2:
                    self.tune_data[name].append('n/a')
                else:
                    self.tune_data[name][2] = 'n/a'

        logger.info("Tune {} result is: {}, Best tune result is: {}".format(trials_count,
                                                                            last_tune_msg,
                                                                            best_tune_msg))
        output_data = [[info_type, 
            '{:.4f} '.format(self.tune_data[info_type][0]) if \
            not isinstance(self.tune_data[info_type][0], str) else self.tune_data[info_type][0], 
            '{:.4f} '.format(self.tune_data[info_type][1]) if \
            not isinstance(self.tune_data[info_type][1], str) else self.tune_data[info_type][1],
            '{:.4f} '.format(self.tune_data[info_type][2]) if \
            not isinstance(self.tune_data[info_type][2], str) else self.tune_data[info_type][2]] \
            for info_type in self.tune_data.keys() if info_type != 'baseline']

        output_data.extend([[obj, 
            '{:.4f} '.format(self.baseline[1][i]) if self.baseline else 'n/a',
            '{:.4f} '.format(self.last_tune_result[1][i]) if self.last_tune_result else 'n/a',
            '{:.4f} '.format(self.best_tune_result[1][i]) if self.best_tune_result else 'n/a'] \
            for i, obj in enumerate(self.objectives.representation)])

        Statistics(output_data,
                   header='Tune Result Statistics',
                   field_names=['Info Type', 'Baseline', 'Tune {} result'.format(trials_count), \
                                                                'Best tune result']).print_stat()


        if self.cfg.tuning.exit_policy.performance_only:
            need_stop = True
        elif timeout == 0 and self.best_tune_result:
            need_stop = True
        elif trials_count >= self.cfg.tuning.exit_policy.max_trials:
            need_stop = True
        else:
            need_stop = False

        return need_stop

    def _save(self):
        """save current tuning state to snapshot for resuming.

        """

        logger.info("Save tuning history to {}.".format(self.history_path))
        with fault_tolerant_file(self.history_path) as f:
            pickle.dump(self, f, protocol=pickle.HIGHEST_PROTOCOL)

    def _find_tuning_history(self, tune_cfg):
        """check if the specified tune_cfg is evaluated or not on same yaml config.

        Args:
            tune_cfg (dict): The tune_cfg to check if evaluated before.

        Returns:
            tuning_history or None: The tuning history containing evaluated tune_cfg.
        """
        for tuning_history in self.tuning_history:
            # only check if a tune_cfg is evaluated under same yam config, excluding
            # some fields in tuning section of yaml, such as tensorboard, snapshot, resume.
            if self._same_yaml(tuning_history['cfg'], self.cfg):
                for history in tuning_history['history']:
                    if history and history['tune_cfg'] == tune_cfg:
                        return tuning_history

        return None

    def _find_history(self, tune_cfg):
        """check if the specified tune_cfg is evaluated or not on same yaml config.

        Returns:
            history or None: The history containing evaluated tune_cfg.
        """
        for tuning_history in self.tuning_history:
            # only check if a tune_cfg is evaluated under same yam config, excluding
            # some fields in tuning section of yaml, such as tensorboard, snapshot, resume.
            if self._same_yaml(tuning_history['cfg'], self.cfg):
                for history in tuning_history['history']:
                    if history and history['tune_cfg'] == tune_cfg:
                        return history
        return None

    def _find_self_tuning_history(self):
        """find self history dict.

        Returns:
            history or None: The history for self.
        """
        for tuning_history in self.tuning_history:
            # only check if a tune_cfg is evaluated under same yam config, excluding
            # some fields in tuning section of yaml, such as tensorboard, snapshot, resume.
            if self._same_yaml(tuning_history['cfg'], self.cfg):
                return tuning_history

        return None

    def _add_tuning_history(self, tune_cfg=None, tune_result=None, **kwargs):
        """add tuning history.
           note this record is added under same yaml config.

        """
        found = False
        d = {'tune_cfg': tune_cfg, 'tune_result': tune_result}
        for tuning_history in self.tuning_history:
            if self._same_yaml(tuning_history['cfg'], self.cfg):
                d.update(kwargs)
                tuning_history['history'].append(d)
                tuning_history['last_tune_result'] = self.last_tune_result
                tuning_history['best_tune_result'] = self.best_tune_result
                tuning_history['cfg'] = self.cfg
                found = True
                break

        if not found:
            tuning_history = {}
            tuning_history['version']  = __version__
            tuning_history['cfg']     = self.cfg
            tuning_history['baseline'] = self.baseline
            tuning_history['last_tune_result'] = self.last_tune_result
            tuning_history['best_tune_result'] = self.best_tune_result
            tuning_history['history']  = []
            if tune_cfg and tune_result:
                d.update(kwargs)
                tuning_history['history'].append(d)
            self.tuning_history.append(tuning_history)

        self._save()

    def _fake_eval_func(self, model):
        return 1.

    def _diagnosis(self):
        import logging
        logger = logging.getLogger()
        iteration_list = self.cfg.tuning.diagnosis.iteration_list
        inspect_type = self.cfg.tuning.diagnosis.inspect_type
        save_to_disk = self.cfg.tuning.diagnosis.save_to_disk
        save_path = self.cfg.tuning.diagnosis.save_path
        inspect_node_lst, updated_cfg = self.adaptor.diagnosis_helper(self._fp32_model, 
                                                                      self.last_qmodel, 
                                                                      self.tune_cfg, 
                                                                      save_path = save_path)
        op_list = self.cfg.tuning.diagnosis.op_list
        if not op_list:
            op_list = list(inspect_node_lst)
        else:
            op_list = list(set(op_list).intersection(inspect_node_lst))

        logger.debug(f'*** Start to inspect tensor :{op_list} in  fp32 model.')
        self.adaptor.inspect_tensor(self._fp32_model,
                                    dataloader=self.calib_dataloader,
                                    op_list=op_list,
                                    iteration_list=iteration_list,
                                    inspect_type=inspect_type, 
                                    save_to_disk=save_to_disk,
                                    save_path= save_path + '/fp32/',
                                    quantization_cfg=updated_cfg)

        logger.debug(f'*** Start to inspect tensor :{op_list} in  quantized model.')
        self.adaptor.inspect_tensor(self.last_qmodel, 
                                    dataloader=self.calib_dataloader,
                                    op_list=op_list,
                                    iteration_list=iteration_list,
                                    inspect_type=inspect_type, 
                                    save_to_disk=save_to_disk,
                                    save_path= save_path + '/quan/',
                                    quantization_cfg=updated_cfg)
    



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import yaml
import logging

from neural_compressor.adaptor.adaptor import adaptor_registry, Adaptor
from neural_compressor.adaptor.query import QueryBackendCapability
from neural_compressor.utils.utility import (LazyImport, GLOBAL_STATE, MODE, CpuInfo,
                                             dump_elapsed_time, singleton)
from neural_compressor.adaptor.mxnet_utils.util import *
from collections import OrderedDict
from ..experimental.data.dataloaders.base_dataloader import BaseDataLoader
from copy import deepcopy
import math

mx = LazyImport("mxnet")
logger = logging.getLogger()


@adaptor_registry
class MxNetAdaptor(Adaptor):
    """The MXNet adaptor layer, do MXNet quantization, calibration, inspect layer tensors.

    Args:
        framework_specific_info (dict): framework specific configuration for quantization.
    """

    def __init__(self, framework_specific_info):
        super(MxNetAdaptor, self).__init__(framework_specific_info)
        assert check_mx_version('1.6.0'), \
            "Need MXNet version >= 1.6.0, but got version: %s" % (mx.__version__)

        self.pre_optimized_model = None
        self.quantizable_nodes = []
        self._qtensor_to_tensor = {}
        self._tensor_to_node = {}
        self.qdataloader = framework_specific_info.get("q_dataloader")
        self.query_handler = MXNetQuery(local_config_file=os.path.join(
            os.path.dirname(__file__), "mxnet.yaml"))

        self.ctx = mx.cpu() if framework_specific_info['device'] == 'cpu' else None
        self.benchmark = (GLOBAL_STATE.STATE == MODE.BENCHMARK)
        assert self.ctx is not None, 'Unsupported device'

    @dump_elapsed_time("Pass quantize model")
    def quantize(self, tune_cfg, nc_model, dataloader, q_func=None):
        """The function is used to do MXNet calibration and quanitization in post-training
           quantization.

        Args:
            tune_cfg (dict):     quantization config.
            nc_model (object): neural_compressor fp32 model to be quantized.
            dataloader (object): calibration dataset.
            q_func (optional):   training function for quantization aware training mode,
                                 unimplement yet for MXNet.

        Returns:
            (MXNetModel): quantized model
        """
        assert q_func is None, "quantization aware training mode is not supported on mxnet"

        calib_cache = nc_model.calib_cache

        def calib_func(tmp_tune_cfg, dataloader):
            quant_cfg, calib_cfg, amp_cfg = parse_tune_config(tmp_tune_cfg, self.quantizable_nodes)
            logger.debug("Dump quantization configurations:")
            logger.debug(quant_cfg)

            sym_model, dataloader = prepare_model_data(nc_model, self.ctx, dataloader)
            qsym_model, calib_tensors = quantize_sym_model(sym_model, self.ctx, quant_cfg)
            calib_data = self._collect_thresholds(sym_model, dataloader, calib_tensors,
                                                  calib_cfg, calib_cache)
            qsym_model = calib_model(qsym_model, calib_data, calib_cfg)
            qsym_model = fuse(qsym_model, self.ctx)  # post-quantization fusion

            if len(amp_cfg['excluded_sym_names']) < len(self.quantizable_nodes):
                qsym_model = amp_convert(qsym_model, dataloader.input_desc, amp_cfg)

            q_nc_model = make_nc_model(nc_model, qsym_model, self.ctx, dataloader.input_desc)
            q_nc_model.calib_cache['last'] = calib_data.th_dict
            q_nc_model.q_config = {
                'mxnet_version': mx.__version__,
                'amp_cfg': amp_cfg,
                'quant_cfg': quant_cfg,
                'calib_cfg': calib_cfg,
                'th_dict': calib_data.th_dict,
                'input_desc': dataloader.input_desc,
                'framework_specific_info': {'device': self.ctx.device_type}}
            return q_nc_model

        calib_sampling_size = tune_cfg.get('calib_sampling_size', 1)
        if isinstance(dataloader, BaseDataLoader):
            batch_size = dataloader.batch_size
            try:
                for i in range(batch_size):
                    if calib_sampling_size % (batch_size - i) == 0:
                        calib_batch_size = batch_size - i
                        if i != 0:
                            logger.warning("Reset `calibration.dataloader.batch_size` field "
                                           "to {}".format(calib_batch_size) +
                                           " to make sure the sampling_size is "
                                           "divisible exactly by batch size")
                        break
                tmp_iterations = int(math.ceil(calib_sampling_size / calib_batch_size))
                tmp_tune_cfg = deepcopy(tune_cfg)
                tmp_tune_cfg['calib_iteration'] = tmp_iterations
                dataloader.batch(calib_batch_size)
                return calib_func(tmp_tune_cfg, dataloader)
            except Exception:  # pragma: no cover
                logger.warning(
                    "Fail to forward with batch size={}, set to {} now.".
                    format(batch_size, 1))
                tmp_tune_cfg = deepcopy(tune_cfg)
                tmp_tune_cfg['calib_iteration'] = calib_sampling_size
                dataloader.batch(1)
                return calib_func(tmp_tune_cfg, dataloader)
        else:
            if hasattr(dataloader, 'batch_size') and \
                    calib_sampling_size % dataloader.batch_size != 0:
                iter = tune_cfg['calib_iteration']
                logger.warning(
                    "Please note that calibration sampling size {} "
                    "isn't divisible exactly by batch size {}. "
                    "So the real sampling size is {}.".
                    format(calib_sampling_size, dataloader.batch_size,
                           dataloader.batch_size * iter))
            return calib_func(tune_cfg, dataloader)

    def _collect_thresholds(self, sym_model, calib_data, calib_tensors, calib_cfg, calib_cache):
        """Calculate thresholds for each tensor. The calibration method can be min/max
           or KL on different tensors.

        Args:
            sym_model (tuple): symbol model (symnet, args, auxs).
            calib_data (DataLoaderWrap): dataset to do calibration on.
            calib_tensors (list): tensors to calibrate
            calib_cfg (dict): calibration config.
            calib_cache (dict): cached calibration results

        Returns:
            (CalibResult): The results of calibration (pair of thresholds for each tensor).
        """
        assert calib_cfg['calib_mode'] == 'naive', \
            '`calib_mode` must be set to `naive`, for `collector.min_max_dict` to be used'

        if calib_cache.get('batches', -1) != calib_cfg['batches']:
            calib_cache['batches'] = calib_cfg['batches']
            calib_cache['kl'] = {}
            calib_cache['minmax'] = {}

        cache_kl = calib_cache['kl']
        cache_minmax = calib_cache['minmax']
        tensors_kl, tensors_minmax = distribute_calib_tensors(calib_tensors, calib_cfg,
                                                              self._tensor_to_node)
        to_collect_kl = tensors_kl - set(cache_kl.keys())
        to_collect_minmax = tensors_minmax - set(cache_minmax.keys())
        collector = CalibCollector(to_collect_kl, to_collect_minmax)

        if len(to_collect_kl) + len(to_collect_minmax) > 0:
            def b_filter():
                for _ in range(calib_cfg['batches']):
                    yield True

            logger.info("Start to collect tensors of the FP32 model.")
            batches = run_forward(sym_model, self.ctx, calib_data, b_filter(),
                                  collector, collector.pre_batch, collector.post_batch)
            logger.info("Get collected tensors of the FP32 model from {} batches.".format(batches))

            if len(collector.include_tensors_kl) > 0:
                cache_kl.update(collector.calc_kl_th_dict(calib_cfg['quantized_dtype']))
            cache_minmax.update(collector.min_max_dict)

        return CalibData(cache_kl, cache_minmax, tensors_kl, tensors_minmax)

    def evaluate(self, nc_model, data_x, postprocess=None,
                 metrics=None, measurer=None, iteration=-1,
                 tensorboard=False, fp32_baseline=False):
        """The function is used to run evaluation on validation dataset.

        Args:
            nc_model (object): model to evaluate.
            data_x (object): data iterator/loader.
            postprocess (object, optional): process the result from the model
            metrics (list): list of evaluate metric.
            measurer (object, optional): for precise benchmark measurement.
            iteration(int, optional): control steps of mini-batch
            tensorboard (boolean, optional): for tensorboard inspect tensor.
            fp32_baseline (boolean, optional): only for compare_label=False pipeline

        Returns:
            acc: evaluate result.
        """
        def b_filter():
            if iteration == -1:
                while True:
                    yield True
            for _ in range(iteration):
                yield True

        def pre_batch(net, batch):
            if measurer is not None:
                measurer.start()

        def post_batch(net, batch, outs):
            if measurer is not None:
                measurer.end()
            _, labels = batch
            outs = ensure_list(outs)
            labels = ensure_list(labels)
            assert len(labels) == len(outs) == 1

            out = outs[0].asnumpy()
            label = labels[0].asnumpy()
            if postprocess is not None:
                out, label = postprocess((out, label))
            if metrics is not None:
                for metric in metrics:
                    metric.update(out, label)

        if isinstance(data_x, BaseDataLoader) and not self.benchmark:
            try:
                sym_model, dataloader = prepare_model_data(nc_model, self.ctx, data_x)
                run_forward(sym_model, self.ctx, dataloader, b_filter(),
                            pre_batch=pre_batch, post_batch=post_batch)
            except Exception:  # pragma: no cover
                logger.warning(
                    "Fail to forward with batch size={}, set to {} now.".
                    format(data_x.batch_size, 1))
                data_x.batch(1)
                sym_model, dataloader = prepare_model_data(nc_model, self.ctx, data_x)
                run_forward(sym_model, self.ctx, dataloader, b_filter(),
                            pre_batch=pre_batch, post_batch=post_batch)
        else:
            sym_model, dataloader = prepare_model_data(nc_model, self.ctx, data_x)
            run_forward(sym_model, self.ctx, dataloader, b_filter(),
                        pre_batch=pre_batch, post_batch=post_batch)
        acc = [metric.result() for metric in metrics] if metrics is not None else 0
        return acc if not isinstance(acc, list) or len(acc) > 1 else acc[0]

    @dump_elapsed_time('Query quantizable operators')
    def query_fw_capability(self, nc_model):
        """Query MXNet quantization capability on the model/op level with the specific model.

        Args:
            nc_model (object): model to query.

        Returns:
            dict: modelwise and opwise config.
        """
        sym_model, self.qdataloader = prepare_model_data(nc_model, self.ctx, self.qdataloader)
        # (TODO) to allign with other fw, set pre_optimized_model here
        self.pre_optimized_model = sym_model

        self.quantizable_nodes, self._tensor_to_node, all_op_nodes = query_quantizable_nodes(
            sym_model, self.ctx, self.qdataloader)

        config = self.query_handler.get_quantization_capability()['int8']
        bf16_config = self.query_handler.get_quantization_capability().get('bf16', {})
        valid_precisions = self.query_handler.get_mixed_precision_combination()
        use_bf16 = ('bf16' in valid_precisions and CpuInfo().bf16) or os.getenv('FORCE_BF16') == '1'
        if use_bf16:
            config = combine_capabilities(config, bf16_config)

        op_type_wise = OrderedDict()
        op_wise = OrderedDict()
        for node in self.quantizable_nodes:
            op_capability = config.get(node['type'], config['default'])
            op_type_wise.setdefault(node['type'], op_capability)
            op_wise.setdefault((node['name'], node['type']), op_capability)

        if use_bf16:
            for node_name, op_name in all_op_nodes.items():
                if (node_name, op_name) not in op_wise and op_name in bf16_config:
                    op_type_wise.setdefault(op_name, bf16_config[op_name])
                    op_wise.setdefault((node_name, op_name), bf16_config[op_name])

        # make sure configurations are independent
        for op, cfg in op_type_wise.items():
            op_type_wise[op] = deepcopy(cfg)
        for key, cfg in op_wise.items():
            op_wise[key] = deepcopy(cfg)

        return {'optypewise': op_type_wise, 'opwise': op_wise}

    def _inspect_tensor(self, nc_model, data_x, node_list=[], iteration_list=[]):
        def b_filter():
            iteration_set = set(iteration_list)
            if len(iteration_set) == 0:
                while True:
                    yield True
            i = 1
            while len(iteration_set) > 0:
                run = (i in iteration_list)
                iteration_set -= {i}
                i += 1
                yield run

        sym_model, dataloader = prepare_model_data(nc_model, self.ctx, data_x)
        collector = TensorCollector(node_list, self._qtensor_to_tensor, self._tensor_to_node)
        num_batches = run_forward(sym_model, self.ctx, dataloader, b_filter(),
                                  collector, pre_batch=collector.pre_batch)
        logger.debug("Inspect batches at {}.".format(num_batches))
        self._qtensor_to_tensor = collector.qtensor_to_tensor
        return collector.tensors_dicts

    def inspect_tensor(self, nc_model, data_x, op_list=[], iteration_list=[],
                       inspect_type='activation', save_to_disk=False,
                       save_path = None, quantization_cfg = None):
        """The function is used by tune strategy class for dumping tensor info.

        Args:
            nc_model (object): The model to do calibration.
            data_x (object): Data iterator/loader.
            op_list (list): list of inspect tensors.
            iteration_list (list): list of inspect iterations.

        Returns:
            dict: includes tensor dicts
        """
        if inspect_type not in ['all', 'activation']:
            raise NotImplementedError()

        tensor_dict_list = self._inspect_tensor(nc_model, data_x, op_list, iteration_list)
        for tensor_dict in tensor_dict_list:
            for key, tensors in tensor_dict.items():
                for tensor_name, (is_quantized, tensor) in tensors.items():
                    tensor_dict[key][tensor_name] = tensor  # discard is_quantized
                    if is_quantized:
                        assert tensor.dtype in QUANTIZATION_DTYPES
                        assert 'last' in nc_model.calib_cache
                        min_th, max_th = nc_model.calib_cache['last'][tensor_name]
                        tensor_dict[key][tensor_name] = mx.nd.contrib.dequantize(
                            tensor,
                            min_range=mx.nd.array([min_th]).squeeze(),
                            max_range=mx.nd.array([max_th]).squeeze(),
                            out_type='float32')
                    tensor_dict[key][tensor_name] = tensor_dict[key][tensor_name].asnumpy()

                # transform to format expected by neural_compressor (assume only 1 tensor for now)
                node = key
                assert len(tensors) == 1, 'Multiple tensors from a single node are not supported'
                tensor = list(tensor_dict[key].values())[0]
                tensor_dict[key] = {node: tensor}

        return {'activation': tensor_dict_list}

    def recover_tuned_model(self, nc_model, q_config):
        """Execute the recover process on the specified model.

            Args:
                tune_cfg (dict): quantization configuration
                nc_model (object): fp32 model
                q_config (dict): recover configuration

            Returns:
                MXNetModel: the quantized model
        """
        if q_config['mxnet_version'] != mx.__version__:  # pragma: no cover
            logger.warning('Attempting to recover a model generated with a different '
                           'version of MXNet ({})'.format(q_config['mxnet_version']))

        sym_model = prepare_model(nc_model, self.ctx, q_config['input_desc'])
        qsym_model, calib_tensors = quantize_sym_model(sym_model, self.ctx, q_config['quant_cfg'])

        calib_data = CalibData()
        calib_data.th_dict = q_config['th_dict']
        assert set(calib_tensors).issubset(calib_data.th_dict.keys())

        qsym_model = calib_model(qsym_model, calib_data, q_config['calib_cfg'])
        qsym_model = fuse(qsym_model, self.ctx)  # post-quantization fusion

        q_nc_model = make_nc_model(nc_model, qsym_model, self.ctx, q_config['input_desc'])
        q_nc_model.calib_cache['last'] = q_config['th_dict']
        q_nc_model.q_config = q_config
        return q_nc_model

    def set_tensor(self, model, tensor_dict):
        '''The function is used by tune strategy class for setting tensor back to model.

           Args:
               model (object): The model to set tensor. Usually it is quantized model.
               tensor_dict (dict): The tensor dict to set. Note the numpy array contains float
                                   value, adaptor layer has the responsibility to quantize to
                                   int8 or int32 to set into the quantized model if needed.
                                   The dict format is something like:
                                   {
                                     'weight0_name': numpy.array,
                                     'bias0_name': numpy.array,
                                     ...
                                   }
        '''
        raise NotImplementedError

    def save(self, model, path):
        model.save(path)


@singleton
class MXNetQuery(QueryBackendCapability):
    def __init__(self, local_config_file):
        super().__init__()
        self.version = mx.__version__
        self.cfg = local_config_file
        self.cur_config = None
        self._one_shot_query()

    def _one_shot_query(self):
        with open(self.cfg) as f:
            content = yaml.safe_load(f)
            try:
                self.cur_config = self._get_specified_version_cfg(content)
            except Exception as e:  # pragma: no cover
                logger.info("Fail to parse {} due to {}.".format(self.cfg, str(e)))
                self.cur_config = None
                raise ValueError("Please check if the format of {} follows Neural Compressor yaml schema.".
                                 format(self.cfg))
        self._update_cfg_with_usr_definition()

    def _update_cfg_with_usr_definition(self):
        from neural_compressor.conf.pythonic_config import mxnet_config
        if mxnet_config.precisions is not None:
            self.cur_config['precisions']['names'] = ','.join(mxnet_config.precisions)

    def _get_specified_version_cfg(self, data):
        """Get the configuration for the current runtime.
        If there's no matched configuration in the input yaml, we'll
        use the `default` field of yaml.

        Args:
            data (Yaml content): input yaml file.

        Returns:
            [dictionary]: the content for specific version.
        """
        default_config = None
        for sub_data in data:
            if sub_data['version']['name'] == self.version:
                return sub_data

            if sub_data['version']['name'] == 'default':
                default_config = sub_data

        return default_config

    def get_version(self):
        """Get the current backend's version string.
        """
        return deepcopy(self.cur_config['version']['name'])

    def get_precisions(self):
        """Get the supported low precisions, e.g ['int8', 'bf16']
        """
        return deepcopy(self.cur_config['precisions']['names'])

    def get_op_types(self):
        """Get the op types for specific backend per low precision.
            e.g {'1.6.0': {'int8': ['Conv2D', 'fully_connected']}}
        """
        return deepcopy(self.cur_config['ops'])

    def get_fuse_patterns(self):
        """Get the fusion patterns for specified op type for every specific precision

        """
        return deepcopy(self.cur_config['patterns'])

    def get_quantization_capability(self):
        """Get the quantization capability of low precision op types.
            e.g, granularity, scheme and etc.

        """
        return deepcopy(self.cur_config['capabilities'])

    def get_mixed_precision_combination(self):
        """Get the valid precision combination base on hardware and user' config.
            e.g['fp32', 'bf16', 'int8']
        """
        if self.cur_config['precisions']['valid_mixed_precisions']:
            return [i.strip() for i in self.cur_config['precisions']['valid_mixed_precisions']]

        return [i.strip() for i in self.get_precisions().split(',')]



# Copyright (c) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ... import globals
from ...utils.line_operation import get_line_indent_level, is_eval_func_model_name, get_line_left_hand_side

import logging

logging.basicConfig(level=globals.logging_level,
                    format='%(asctime)s %(levelname)s %(message)s',
                    datefmt='%a, %d %b %Y %H:%M:%S +0000')
logger = logging.getLogger(__name__)


class DummyDataLoader(object):
    def __init__(self, list_model_def_instance):
        self.list_model_def_instance = list_model_def_instance

    def print_info(self):
        for i in self.list_model_def_instance:
            logger.debug(f"i.print_info(): {i.print_info()}")

    # collect file transformation info and register (store) in globals 
    # (i.e. which file to add which lines at which location)
    def register_transformation(self): 
        list_code = []
        for i in globals.list_code_path:
            list_code.append(open(i, 'r').read())

        for ins in self.list_model_def_instance:
            model_name = ins.model_name
            file_path = ins.file_path
            model_def_line_idx = ins.model_def_line_idx
            function_def_line_idx = ins.function_def_line_idx
            class_name = ins.class_name

            # transformation
            file_path_idx = globals.list_code_path.index(file_path)
            lines = list_code[file_path_idx].split('\n')
            line_idx = 0

            # search DataLoader
            dataloader_name = ""
            for i in range(len(lines)): # each item is a str of this code line
                line = lines[i]
                if "DataLoader(" in line and "=" in line and line.find("=") < line.find("DataLoader"):
                    dataloader_name = get_line_left_hand_side(line)
                    dataloader_def_line_idx = i

            if dataloader_name != "":
                return
            else:
                input_dimension_str = "3, 224, 224)"
                for i in range(len(lines)):
                    line = lines[i]
                    if ("input" in line and "=" in line and \
                        line.find("=") > line.find("input")) or \
                            ("image" in line and "=" in line and line.find("=") > line.find("image")):
                        input_dimension_str = line[line.find(",")+2:]

                for i in range(len(lines)):
                    line = lines[i]
                    if line_idx == model_def_line_idx:
                        indent_level = get_line_indent_level(line)
                        lines_to_insert = ""
                        lines_to_insert += " " * indent_level + "import torch" + "\n"
                        lines_to_insert += " " * indent_level + "from torch.utils.data import Dataset" + "\n"
                        lines_to_insert += " " * indent_level + "class DummyDataset(Dataset):" + "\n"
                        lines_to_insert += " " * indent_level + \
                            "    def __init__(self, *shapes, num_samples: int = 10000):" + "\n"
                        lines_to_insert += " " * indent_level + "        super().__init__()" + "\n"
                        lines_to_insert += " " * indent_level + "        self.shapes = shapes" + "\n"
                        lines_to_insert += " " * indent_level + "        self.num_samples = num_samples" + "\n"
                        lines_to_insert += " " * indent_level + "    def __len__(self):" + "\n"
                        lines_to_insert += " " * indent_level + "        return self.num_samples" + "\n"
                        lines_to_insert += " " * indent_level + "    def __getitem__(self, idx: int):" + "\n"
                        lines_to_insert += " " * indent_level + "        sample = []" + "\n"
                        lines_to_insert += " " * indent_level + "        for shape in self.shapes:" + "\n"
                        lines_to_insert += " " * indent_level + "            spl = torch.rand(*shape)" + "\n"
                        lines_to_insert += " " * indent_level + "            sample.append(spl)" + "\n"
                        lines_to_insert += " " * indent_level + "        return sample" + "\n"
                        lines_to_insert += " " * indent_level + "from torch.utils.data import DataLoader" + "\n"
                        lines_to_insert += " " * indent_level + \
                            "dummy_dataset = DummyDataset((" + input_dimension_str + ", (1, ))" + "\n"
                        lines_to_insert += " " * indent_level + \
                            "dummy_dataloader = DataLoader(dummy_dataset, batch_size=1)"

                        trans_insert_location = 0

                        if file_path not in globals.list_trans_insert_modified_file:
                            globals.list_trans_insert_modified_file.append(file_path)
                            globals.list_trans_insert_location_idxs.append([trans_insert_location])
                            globals.list_trans_insert_number_insert_lines.append(
                                [lines_to_insert.count("\n") + 1]
                            )
                            globals.list_trans_insert_lines_to_insert.append([lines_to_insert])
                        else:
                            idx = globals.list_trans_insert_modified_file.index(file_path)
                            globals.list_trans_insert_location_idxs[idx].append(trans_insert_location)
                            globals.list_trans_insert_number_insert_lines[idx].append(
                                lines_to_insert.count("\n") + 1
                            )
                            globals.list_trans_insert_lines_to_insert[idx].append(lines_to_insert)

                    line_idx += 1

        logger.debug(f"globals.list_trans_insert_modified_file: {globals.list_trans_insert_modified_file}")
        logger.debug(f"globals.list_trans_insert_location_idxs: {globals.list_trans_insert_location_idxs}")
        logger.debug(f"globals.list_trans_insert_number_insert_lines: {globals.list_trans_insert_number_insert_lines}")
        logger.debug(f"globals.list_trans_insert_lines_to_insert: {globals.list_trans_insert_lines_to_insert}")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import ast
import random
from typing import List, Tuple

import numpy as np
import pandas as pd
from sklearn.model_selection import train_test_split

from neural_compressor.utils import logger


class ParameterManager:
    """The ParameterManager class handles the super-network encoding representation and translates between
    the 1-hot predictor, pymoo, and super-network dictionary formats during search.
    """

    def __init__(
        self,
        param_dict: dict,
        verbose: bool = False,
        seed: int = 0,
    ) -> None:
        self.param_dict = param_dict
        self.verbose = verbose
        self.mapper, self.param_upperbound, self.param_count = self.process_param_dict()
        self.inv_mapper = self._inv_mapper()
        self.set_seed(seed)

    def process_param_dict(self) -> Tuple[list, list, int]:
        '''
        Builds a parameter mapping arrays and an upper-bound vector for pymoo.
        '''
        parameter_count = 0
        parameter_bound = list()
        parameter_upperbound = list()
        parameter_mapper = list()

        for parameter, options in self.param_dict.items():
            # How many variables should be searched for
            parameter_count += options['count']
            parameter_bound.append(options['count'])

            # How many variables for each parameter
            for i in range(options['count']):
                parameter_upperbound.append(len(options['vars']) - 1)
                index_simple = [x for x in range(len(options['vars']))]
                parameter_mapper.append(dict(zip(index_simple, options['vars'])))

        if self.verbose: # pragma: no cover
            logger.info(
                '[DyNAS-T] Problem definition variables: {}'.format(parameter_count)
            )
            logger.info(
                '[DyNAS-T] Variable Upper Bound array: {}'.format(parameter_upperbound)
            )
            logger.info(
                '[DyNAS-T] Mapping dictionary created of length: {}'.format(
                    len(parameter_mapper)
                )
            )
            logger.info('[DyNAS-T] Parameter Bound: {}'.format(parameter_bound))

        return parameter_mapper, parameter_upperbound, parameter_count

    def _inv_mapper(self) -> list:
        '''
        Builds inverse of self.mapper
        '''
        inv_parameter_mapper = list()

        for value in self.mapper:
            inverse_dict = {v: k for k, v in value.items()}
            inv_parameter_mapper.append(inverse_dict)

        return inv_parameter_mapper

    def onehot_generic(self, in_array: list) -> np.ndarray:
        '''
        This is a generic approach to one-hot vectorization for predictor training
        and testing. It does not account for unused parameter mapping (e.g. block depth).
        For unused parameter mapping, the end user will need to provide a custom solution.

        input_array - the pymoo individual 1-D vector
        mapper - the map for elastic parameters of the supernetwork
        '''
        # Insure compatible array and mapper
        assert len(in_array) == len(self.mapper)

        onehot = list()

        # This function converts a pymoo input vector to a one-hot feature vector
        for i in range(len(self.mapper)):
            segment = [0 for _ in range(len(self.mapper[i]))]
            segment[in_array[i]] = 1
            onehot.extend(segment)

        return np.array(onehot)

    def random_sample(self) -> list:
        '''
        Generates a random subnetwork from the possible elastic parameter range
        '''
        pymoo_vector = list()
        for i in range(len(self.mapper)):
            options = [x for x in range(len(self.mapper[i]))]
            pymoo_vector.append(random.choice(options))

        return pymoo_vector

    def random_samples(self, size: int = 100, trial_limit: int = 100000) -> List[list]:
        '''
        Generates a list of random subnetworks from the possible elastic parameter range
        '''
        pymoo_vector_list = list()

        trials = 0
        while len(pymoo_vector_list) < size and trials < trial_limit:
            sample = self.random_sample()
            if sample not in pymoo_vector_list:
                pymoo_vector_list.append(sample)
            trials += 1

        if trials >= trial_limit:
            logger.warning('[DyNAS-T] Unable to create unique list of samples.')

        return pymoo_vector_list

    def translate2param(self, pymoo_vector: list) -> dict:
        '''
        Translate a PyMoo 1-D parameter vector back to the elastic parameter dictionary format
        '''
        output = dict()

        # Assign (and map) each vector element to the appropriate parameter dictionary key
        counter = 0
        for key, value in self.param_dict.items():
            output[key] = list()
            for i in range(value['count']):
                output[key].append(self.mapper[counter][pymoo_vector[counter]])
                counter += 1

        # Insure correct vector mapping occurred
        assert counter == len(self.mapper)

        return output

    def translate2pymoo(self, parameters: dict) -> list:
        '''
        Translate a single parameter dict to pymoo vector
        '''
        output = list()

        mapper_counter = 0
        for key, value in self.param_dict.items():
            param_counter = 0
            for i in range(value['count']):
                output.append(
                    self.inv_mapper[mapper_counter][parameters[key][param_counter]]
                )
                mapper_counter += 1
                param_counter += 1

        return output

    def import_csv(
        self,
        filepath: str,
        config: str,
        objective: str,
        column_names: List[str] = None,
        drop_duplicates: bool = True,
    ) -> pd.DataFrame:
        '''
        Import a csv file generated from a supernetwork search for the purpose
        of training a predictor.

        filepath - path of the csv to be imported.
        config - the subnetwork configuration
        objective - target/label for the subnet configuration (e.g. accuracy, latency)
        column_names - a list of column names for the dataframe
        df - the output dataframe that contains the original config dict, pymoo, and 1-hot
             equivalent vector for training.
        '''

        if column_names == None:
            df = pd.read_csv(filepath)
        else:
            df = pd.read_csv(filepath)
            df.columns = column_names
        df = df[[config, objective]]

        # OFA corner case coverage
        df[config] = df[config].replace({'null': 'None'}, regex=True)

        if drop_duplicates:
            df.drop_duplicates(subset=[config], inplace=True)
            df.reset_index(drop=True, inplace=True)

        convert_to_dict = list()
        convert_to_pymoo = list()
        convert_to_onehot = list()
        for i in range(len(df)):
            # Elastic Param Config format
            config_as_dict = ast.literal_eval(df[config].iloc[i])
            convert_to_dict.append(config_as_dict)
            # PyMoo 1-D vector format
            config_as_pymoo = self.translate2pymoo(config_as_dict)
            convert_to_pymoo.append(config_as_pymoo)
            # Onehot preditor format
            config_as_onehot = self.onehot_generic(config_as_pymoo)
            convert_to_onehot.append(config_as_onehot)

        df[config] = convert_to_dict
        df['config_pymoo'] = convert_to_pymoo
        df['config_onehot'] = convert_to_onehot

        return df

    def set_seed(self, seed) -> None:
        '''
        Set the random seed for randomized subnet generation and test/train split
        '''
        self.seed = seed
        random.seed(seed)

    @staticmethod
    def create_training_set(
        dataframe: pd.DataFrame,
        train_with_all: bool = True,
        split: float = 0.33,
        seed: bool = None,
    ) -> Tuple[list, list, list, list]:
        '''
        Create a sklearn compatible test/train set from an imported results csv
        after "import_csv" method is run.
        '''

        collect_rows = list()
        for i in range(len(dataframe)):
            collect_rows.append(np.asarray(dataframe['config_onehot'].iloc[i]))
        features = np.asarray(collect_rows)

        labels = dataframe.drop(
            columns=['config', 'config_pymoo', 'config_onehot']
        ).values

        assert len(features) == len(labels)

        if train_with_all:
            logger.info('[DyNAS-T] Training set size={}'.format(len(labels)))
            return features, labels
        else:
            features_train, features_test, labels_train, labels_test = train_test_split(
                features, labels, test_size=split, random_state=seed
            )
            logger.info(
                '[DyNAS-T] Test ({}) Train ({}) ratio is {}.'.format(
                    len(labels_train), len(labels_test), split
                )
            )
            return features_train, features_test, labels_train, labels_test
